import json

import pytest
from fastapi.testclient import TestClient

from ragna.assistants import RagnaDemoAssistant
from ragna.deploy import Config
from ragna.deploy._api import app

from .utils import authenticate


class TestAssistant(RagnaDemoAssistant):
    def answer(self, prompt, sources, *, multiple_answer_chunks: bool):
        content = next(super().answer(prompt, sources))

        if multiple_answer_chunks:
            for chunk in content.split(" "):
                yield f"{chunk} "
        else:
            yield content


@pytest.mark.parametrize("multiple_answer_chunks", [True, False])
@pytest.mark.parametrize("stream_answer", [True, False])
def test_e2e(tmp_local_root, multiple_answer_chunks, stream_answer):
    config = Config(local_root=tmp_local_root, assistants=[TestAssistant])

    document_root = config.local_root / "documents"
    document_root.mkdir()
    document_path = document_root / "test.txt"
    with open(document_path, "w") as file:
        file.write("!\n")

    with TestClient(app(config=config, ignore_unavailable_components=False)) as client:
        authenticate(client)

        assert client.get("/chats").raise_for_status().json() == []

        document_upload = (
            client.post("/document", json={"name": document_path.name})
            .raise_for_status()
            .json()
        )
        document = document_upload["document"]
        assert document["name"] == document_path.name

        parameters = document_upload["parameters"]
        with open(document_path, "rb") as file:
            client.request(
                parameters["method"],
                parameters["url"],
                data=parameters["data"],
                files={"file": file},
            )

        components = client.get("/components").raise_for_status().json()
        documents = components["documents"]
        assert set(documents) == config.document.supported_suffixes()
        source_storages = [
            json_schema["title"] for json_schema in components["source_storages"]
        ]
        assert source_storages == [
            source_storage.display_name() for source_storage in config.source_storages
        ]
        assistants = [json_schema["title"] for json_schema in components["assistants"]]
        assert assistants == [
            assistant.display_name() for assistant in config.assistants
        ]

        source_storage = source_storages[0]
        assistant = assistants[0]

        chat_metadata = {
            "name": "test-chat",
            "source_storage": source_storage,
            "assistant": assistant,
            "params": {"multiple_answer_chunks": multiple_answer_chunks},
            "documents": [document],
        }
        chat = client.post("/chats", json=chat_metadata).raise_for_status().json()
        assert chat["metadata"] == chat_metadata
        assert not chat["prepared"]
        assert chat["messages"] == []

        assert client.get("/chats").raise_for_status().json() == [chat]
        assert client.get(f"/chats/{chat['id']}").raise_for_status().json() == chat

        message = client.post(f"/chats/{chat['id']}/prepare").raise_for_status().json()
        assert message["role"] == "system"
        assert message["sources"] == []

        chat = client.get(f"/chats/{chat['id']}").raise_for_status().json()
        assert chat["prepared"]
        assert len(chat["messages"]) == 1
        assert chat["messages"][-1] == message

        prompt = "?"
        if stream_answer:
            with client.stream(
                "POST",
                f"/chats/{chat['id']}/answer",
                json={"prompt": prompt, "stream": True},
            ) as response:
                chunks = [json.loads(chunk) for chunk in response.iter_lines()]
            message = chunks[0]
            assert all(chunk["sources"] is None for chunk in chunks[1:])
            message["content"] = "".join(chunk["content"] for chunk in chunks)
        else:
            message = (
                client.post(f"/chats/{chat['id']}/answer", json={"prompt": prompt})
                .raise_for_status()
                .json()
            )

        assert message["role"] == "assistant"
        assert {source["document"]["name"] for source in message["sources"]} == {
            document_path.name
        }

        chat = client.get(f"/chats/{chat['id']}").raise_for_status().json()
        assert len(chat["messages"]) == 3
        assert (
            chat["messages"][-2]["role"] == "user"
            and chat["messages"][-2]["sources"] == []
            and chat["messages"][-2]["content"] == prompt
        )
        assert chat["messages"][-1] == message

        client.delete(f"/chats/{chat['id']}").raise_for_status()
        assert client.get("/chats").raise_for_status().json() == []



from __future__ import annotations

import itertools
from pathlib import Path
from typing import Annotated, Any, Callable, Generic, Type, TypeVar, Union

import tomlkit
import tomlkit.container
import tomlkit.items
from pydantic import (
    AfterValidator,
    Field,
    ImportString,
    model_validator,
)
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)

import ragna
from ragna._utils import make_directory
from ragna.core import Assistant, Document, RagnaException, SourceStorage

from ._authentication import Authentication

T = TypeVar("T")


class AfterConfigValidateDefault(Generic[T]):
    """This class exists for a specific use case:

    - We have values for which we need the validated config to compute the default,
      e.g. the API default origins can only be computed after we know the UI hostname
      and port.
    - We want to use the plain annotations rather than allowing a sentinel type, e.g.
      `str` vs. `Optional[str]`.
    """

    def __init__(self, make_default: Callable[[Config], T]) -> None:
        self.make_default = make_default

    @classmethod
    def make(cls, make_default: Callable[[Config], T]) -> Any:
        """Creates a default sentinel that is resolved after the config is validated.

        Args:
            make_default: Callable that takes the validated config and returns the
                resolved value.
        """
        return Field(default=cls(make_default), validate_default=False)


class ConfigBase(BaseSettings):
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        # This order is needed to prioritize values from environment variables over
        # values from a configuration file. However, since the config instantiation from
        # a config file goes through the regular constructor of the Python object, we
        # are also implicitly prioritizing environment variables over values passed
        # explicitly passed to the constructor. For example, if the environment variable
        # 'RAGNA_LOCAL_ROOT' is set, any values passed to `Config(local_root=...)` are
        # ignored.
        # FIXME: Find a way to achieve the following priorities:
        #  1. Explicitly passed to Python object
        #  2. Environment variable
        #  3. Configuration file
        #  4. Default
        return env_settings, init_settings

    def _resolve_default_sentinels(self, config: Config) -> None:
        for name, info in self.model_fields.items():
            value = getattr(self, name)
            if isinstance(value, ConfigBase):
                value._resolve_default_sentinels(config)
            elif isinstance(value, AfterConfigValidateDefault):
                setattr(self, name, value.make_default(config))

    def __str__(self) -> str:
        toml = tomlkit.item(self.model_dump(mode="json"))
        self._set_multiline_array(toml)
        return toml.as_string()

    def _set_multiline_array(self, item: tomlkit.items.Item) -> None:
        if isinstance(item, tomlkit.items.Array):
            item.multiline(True)

        if not isinstance(item, tomlkit.items.Table):
            return

        container = item.value
        for child in itertools.chain(
            (value for _, value in container.body), container.value.values()
        ):
            self._set_multiline_array(child)


def make_default_origins(config: Config) -> list[str]:
    return [f"http://{config.ui.hostname}:{config.ui.port}"]


class ApiConfig(ConfigBase):
    model_config = SettingsConfigDict(env_prefix="ragna_api_")

    hostname: str = "127.0.0.1"
    port: int = 31476
    root_path: str = ""
    url: str = AfterConfigValidateDefault.make(
        lambda config: f"http://{config.api.hostname}:{config.api.port}{config.api.root_path}",
    )
    database_url: str = AfterConfigValidateDefault.make(
        lambda config: f"sqlite:///{config.local_root}/ragna.db",
    )
    origins: list[str] = AfterConfigValidateDefault.make(make_default_origins)


class UiConfig(ConfigBase):
    model_config = SettingsConfigDict(env_prefix="ragna_ui_")

    hostname: str = "127.0.0.1"
    port: int = 31477
    origins: list[str] = AfterConfigValidateDefault.make(make_default_origins)


class Config(ConfigBase):
    """Ragna configuration"""

    model_config = SettingsConfigDict(env_prefix="ragna_")

    local_root: Annotated[Path, AfterValidator(make_directory)] = Field(
        default_factory=ragna.local_root
    )

    authentication: ImportString[
        type[Authentication]
    ] = "ragna.deploy.RagnaDemoAuthentication"  # type: ignore[assignment]

    document: ImportString[type[Document]] = "ragna.core.LocalDocument"  # type: ignore[assignment]
    source_storages: list[ImportString[type[SourceStorage]]] = [
        "ragna.source_storages.RagnaDemoSourceStorage"  # type: ignore[list-item]
    ]
    assistants: list[ImportString[type[Assistant]]] = [
        "ragna.assistants.RagnaDemoAssistant"  # type: ignore[list-item]
    ]

    api: ApiConfig = Field(default_factory=ApiConfig)
    ui: UiConfig = Field(default_factory=UiConfig)

    @model_validator(mode="after")
    def _validate_model(self) -> Config:
        self._resolve_default_sentinels(self)
        return self

    @classmethod
    def from_file(cls, path: Union[str, Path]) -> Config:
        path = Path(path).expanduser().resolve()
        if not path.is_file():
            raise RagnaException(f"{path} does not exist.")

        with open(path) as file:
            return cls.model_validate(tomlkit.load(file).unwrap())

    def to_file(self, path: Union[str, Path], *, force: bool = False) -> None:
        path = Path(path).expanduser().resolve()
        if path.exists() and not force:
            raise RagnaException(f"{path} already exists.")

        with open(path, "w") as file:
            file.write(str(self))



import importlib
from pathlib import Path

HERE = Path(__file__).parent
PROJECT_ROOT = HERE.parent
PACKAGE_ROOT = PROJECT_ROOT / "ragna"


def main():
    for path in PACKAGE_ROOT.rglob("*.py"):
        if path.name == "__init__.py":
            path = path.parent
        else:
            path = path.with_suffix("")

        path = path.relative_to(PROJECT_ROOT)

        if any(part.startswith("_") for part in path.parts):
            continue

        name = path.as_posix().replace("/", ".")

        try:
            importlib.import_module(name)
        except Exception as exc:
            raise ImportError(
                f"Trying to import '{name}' raise the error above"
            ) from exc
        else:
            print(name)


if __name__ == "__main__":
    main()



import asyncio

from ragna import Rag
from ragna.assistants import RagnaDemoAssistant
from ragna.core import LocalDocument, Message, MessageRole
from ragna.source_storages import RagnaDemoSourceStorage


def test_e2e(tmp_local_root):
    document_root = tmp_local_root / "documents"
    document_root.mkdir()
    document_path = document_root / "test.txt"
    with open(document_path, "w") as file:
        file.write("!\n")

    document = LocalDocument.from_path(document_path)

    async def main(*, documents, source_storage, assistant):
        async with Rag().chat(
            documents=documents,
            source_storage=source_storage,
            assistant=assistant,
        ) as chat:
            return await chat.answer("?")

    answer = asyncio.run(
        main(
            documents=[document],
            source_storage=RagnaDemoSourceStorage,
            assistant=RagnaDemoAssistant,
        )
    )

    assert isinstance(answer, Message)
    assert answer.role is MessageRole.ASSISTANT
    assert {source.document.name for source in answer.sources} == {document.name}



import json
from datetime import datetime

import emoji
import httpx
import param


class RagnaAuthTokenExpiredException(Exception):
    """Just a wrapper around Exception"""

    pass


# The goal is this class is to provide ready-to-use functions to interact with the API
class ApiWrapper(param.Parameterized):
    auth_token = param.String(default=None)

    def __init__(self, api_url, **params):
        self.client = httpx.AsyncClient(base_url=api_url, timeout=60)

        super().__init__(**params)

        try:
            # If no auth token is provided, we use the API base URL and only test the API is up.
            # else, we test the API is up *and* the token is valid.
            endpoint = (
                api_url + "/components" if self.auth_token is not None else api_url
            )
            httpx.get(
                endpoint, headers={"Authorization": f"Bearer {self.auth_token}"}
            ).raise_for_status()

        except httpx.HTTPStatusError as e:
            # unauthorized - the token is invalid
            if e.response.status_code == 401:
                raise RagnaAuthTokenExpiredException("Unauthorized")
            else:
                raise e

    async def auth(self, username, password):
        self.auth_token = (
            (
                await self.client.post(
                    "/token",
                    data={"username": username, "password": password},
                )
            )
            .raise_for_status()
            .json()
        )

        return True

    @param.depends("auth_token", watch=True, on_init=True)
    def update_auth_header(self):
        self.client.headers["Authorization"] = f"Bearer {self.auth_token}"

    async def get_chats(self):
        json_data = (await self.client.get("/chats")).raise_for_status().json()
        for chat in json_data:
            chat["messages"] = [self.improve_message(msg) for msg in chat["messages"]]
        return json_data

    async def answer(self, chat_id, prompt):
        async with self.client.stream(
            "POST",
            f"/chats/{chat_id}/answer",
            json={"prompt": prompt, "stream": True},
        ) as response:
            async for data in response.aiter_lines():
                yield self.improve_message(json.loads(data))

    async def get_components(self):
        return (await self.client.get("/components")).raise_for_status().json()

    # Upload and related functions
    def upload_endpoints(self):
        return {
            "informations_endpoint": f"{self.client.base_url}/document",
        }

    async def start_and_prepare(
        self, name, documents, source_storage, assistant, params
    ):
        response = await self.client.post(
            "/chats",
            json={
                "name": name,
                "documents": documents,
                "source_storage": source_storage,
                "assistant": assistant,
                "params": params,
            },
        )
        chat = response.raise_for_status().json()

        response = await self.client.post(f"/chats/{chat['id']}/prepare", timeout=None)
        response.raise_for_status()

        return chat["id"]

    def improve_message(self, msg):
        msg["timestamp"] = datetime.strptime(msg["timestamp"], "%Y-%m-%dT%H:%M:%S.%f")
        msg["content"] = emoji.emojize(msg["content"], language="alias")
        return msg



"""
UI Helpers
"""
from typing import Iterable, Optional, Union

import panel as pn


def divider():
    return pn.layout.Divider(styles={"padding": "0em 1em"})


def apply_design_modifiers():
    apply_design_modifiers_source_accordion()
    # add here calls to other design modifiers,
    #   group them per UI component


def apply_design_modifiers_source_accordion():
    pn.theme.fast.Fast.modifiers[pn.layout.Accordion] = {
        "stylesheets": [
            """ :host { 
                                    height: 100%
                                    }
                           """
        ]
    }

    pn.theme.fast.Fast.modifiers[pn.layout.Card] = {
        "stylesheets": [
            """ 

                        /* Define some variables */
                        :host {
                            --ragna-accordion-header-height: 50px;
                        } 

                        /* Resets some existing styles */
                        :host(.accordion) { 
                            margin-left: 0px;
                            margin-top: 0px;
                            outline: none;
                        }
                        
                        /* Styles the button itself */
                        button.accordion-header { 
                            background-color: white !important;
                            height: var(--ragna-accordion-header-height);
                            padding-top: 0px;
                            padding-bottom: 0px;
                            outline:0px;
                            margin-left: 15px;
                            margin-right: 15px;
                            width: calc(100% - 30px);
                            border-bottom: 2px solid #D9D9D9;
                        
                        }
                        
                        button.accordion-header div.card-button {
                            font-size: 11px;
                            padding-top: 5px;
                            margin-left: 0px;
                            margin-right: 10px;
                        }
                        
                        div.card-header-row {
                            height: var(--ragna-accordion-header-height);
                            background-color: unset !important;
                        }

                        /* styles the content of the sources content (the expanding areas of the Accordion) */
                        div.bk-panel-models-markup-HTML.markdown {
                            margin-left: 15px;
                            margin-right: 15px;
                            margin-top:0px;
                        }

                    """
        ]
    }

    pn.theme.fast.Fast.modifiers[pn.pane.HTML] = {
        "stylesheets": [
            """ :host(.card-title) {
                                height: var(--ragna-accordion-header-height);
                                margin: 0px;
                            }
                        
                            :host(.card-title) div {
                                height: var(--ragna-accordion-header-height);
                                
                                display:flex;
                                flex-direction:row;
                                align-items:center;
                            }
                        

                            :host(.card-title) h3 {
                                font-weight: normal;
                            }
                        """
        ]
    }

    pn.theme.fast.Fast.modifiers[pn.pane.Markdown] = {
        "stylesheets": [
            """  /* Styles the content of the sources content (the expanding areas of the Accordion).
                            This fixes a small margin-top that is added by default and that leads to overflowing content 
                            in some cases.
                            */
                            :host(.source-content) p:nth-of-type(1)  {
                                margin-top: 0px;
                            }
                        """
        ]
    }


"""
CSS constants
"""


def stylesheets(
    *class_selectors: tuple[Union[str, Iterable[str]], dict[str, str]]
) -> Optional[list[str]]:
    if not class_selectors:
        return None

    return [
        "\n".join(
            [
                f"{selector if isinstance(selector, str) else ', '.join(selector)} {{",
                *[
                    f"    {property}: {value};"
                    for property, value in declarations.items()
                ],
                "}",
            ]
        )
        for selector, declarations in class_selectors
    ]


MAIN_COLOR = "#DF5538"  # "rgba(223, 85, 56, 1)"


# MAIN_COLOR_LIGHT = "#10BBE580"
# MAIN_COLOR_LIGHTER = "#E1E8E3"
# TABS_SIDEBAR_BKGROUND_COLOR = "#EAEAEA"
# TABS_SIDEBAR_WIDTH = "20em"

# set modal height
CONFIG_MODAL_MIN_HEIGHT = 610
CONFIG_MODAL_MAX_HEIGHT = 850
CONFIG_MODAL_WIDTH = 800

WELCOME_MODAL_HEIGHT = 275
WELCOME_MODAL_WIDTH = 530


APP_RAW = """

:root {
    --body-font: 'Inter', sans-serif !important;
    --accent-color: {{MAIN_COLOR}} !important;
}

* {
    font-family: 'Inter', sans-serif;
}

.main {
    padding: 0px !important;

}

.pn-wrapper {
    padding: 0px;
}

div.card-margin {
    margin: 0px !important;
    height: 100% !important;
}


/* Hide the whole header */

#header {
    display: none;
}


div#content {
    height: calc(100vh);
}

/* Fix the size of the modal */
#pn-Modal {
    --dialog-width: 800px !important;
    --dialog-height:500px !important; 
}

/* Hide the default close button of the modal */
.pn-modal-close {
    display: none !important;
}


/* Hide the fullscreen button of the template */ 
span.fullscreen-button {
    display:none;
}

""".replace("{{MAIN_COLOR}}", MAIN_COLOR)


CHAT_INTERFACE_CUSTOM_BUTTON = """
:host(.solid) .bk-btn.bk-btn-default {
    background-color: transparent;
    color: gray;
}


.bk-btn {
    border-radius: 0;
    padding: 0;
    font-size: 14px;
}
    """

BK_INPUT_GRAY_BORDER = (
    """ .bk-input {border-color: var(--neutral-color) !important;} """
)


SS_LABEL_STYLE = """
:host {
    margin-top:20px;
}

label, .bk-slider-title {
    font-weight: bold;
    margin-bottom: 5px;
}

.bk-slider-value {
    font-weight: normal;
}

.noUi-target {
    border-color: var(--accent-fill-active);
}

.noUi-handle {
    border-radius: 50%;
}

.noUi-handle, .noUi-connects {
    background-color: var(--clear-button-active);
}

.noUi-target[disabled='true'] {
    border-color: darkgray;
}


.noUi-target[disabled='true'] .noUi-connect {
    background-color: darkgray !important;
}


:host(.disabled) div.bk-slider-title {
    color: darkgray !important;
}


"""


SS_ADVANCED_UI_CARD = """
:host {
    border : none !important;
    outline: none !important;
    background-color: unset !important;
}
"""


message_loading_indicator = f""" 
            <div style="height:48px;">
                <svg xmlns="http://www.w3.org/2000/svg" width="48" height="48" viewBox="0 0 24 24">
                    <circle cx="18" cy="12" r="0" fill="{MAIN_COLOR}">
                        <animate attributeName="r" begin=".67" calcMode="spline" dur="1.5s" keySplines="0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8" repeatCount="indefinite" values="0;2;0;0"/>
                    </circle>
                    <circle cx="12" cy="12" r="0" fill="{MAIN_COLOR}">
                        <animate attributeName="r" begin=".33" calcMode="spline" dur="1.5s" keySplines="0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8" repeatCount="indefinite" values="0;2;0;0"/>
                    </circle>
                    <circle cx="6" cy="12" r="0" fill="{MAIN_COLOR}">
                        <animate attributeName="r" begin="0" calcMode="spline" dur="1.5s" keySplines="0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8;0.2 0.2 0.4 0.8" repeatCount="indefinite" values="0;2;0;0"/>
                    </circle>
                </svg>
            </div>
            """



"""
# Python API

The [Python API](../../references/python-api.md) is the best place to get started with
Ragna and understand its key components. It's also the best way to continue
experimenting with components and configurations for your particular use case.

This tutorial walks you through basic steps of using Ragnas Python API.
"""

# %%
# Before we start this tutorial, we import some helpers.

import sys
from pathlib import Path

sys.path.insert(0, str(Path.cwd().parent))

import documentation_helpers

# %%
# ## Step 1: Select relevant documents
#
# Ragna uses the RAG technique to answer questions. The context in which the questions
# will be answered comes from documents that you provide. For this tutorial, let's use a
# sample document that includes some information about Ragna.

document_path = documentation_helpers.assets / "ragna.txt"

with open(document_path) as file:
    print(file.read())

# %%
# !!! tip
#
#     Ragna supports the following document types:
#
#     - [`.txt`][ragna.core.PlainTextDocumentHandler]
#     - [`.md`][ragna.core.PlainTextDocumentHandler]
#     - [`.pdf`][ragna.core.PdfDocumentHandler]
#     - [`.docx`][ragna.core.DocxDocumentHandler]
#     - [`.pptx`][ragna.core.PptxDocumentHandler]

# %%
# ## Step 2: Select a source storage
#
# To effectively retrieve the relevant content of the documents, it needs to be stored
# in a [`SourceStorage`][ragna.core.SourceStorage]. In a regular use case this is a
# vector database, but any database with text search capabilities can be used. For this
# tutorial, we are going to use a demo source storage for simplicity.

from ragna.source_storages import RagnaDemoSourceStorage

# %%
# !!! tip
#
#     Ragna has builtin support for the following source storages:
#
#     - [ragna.source_storages.Chroma][]
#     - [ragna.source_storages.LanceDB][]

# %%
# ## Step 3: Select an assistant
#
# Now that we have a way to retrieve relevant sources for a given user prompt, we now
# need something to actually provide an answer. This is performed by an
# [`Assistant`][ragna.core.Assistant], which is Ragnas abstraction around Large Language
# Models (LLMs). For this tutorial, we are going to use a demo assistant for simplicity.

from ragna.assistants import RagnaDemoAssistant

# %%
# !!! tip
#
#     Ragna has builtin support for the following assistants:
#
#     - [Anthropic](https://www.anthropic.com/)
#       - [ragna.assistants.Claude][]
#       - [ragna.assistants.ClaudeInstant][]
#     - [Cohere](https://cohere.com/)
#       - [ragna.assistants.Command][]
#       - [ragna.assistants.CommandLight][]
#     - [Google](https://ai.google.dev/)
#       - [ragna.assistants.GeminiPro][]
#       - [ragna.assistants.GeminiUltra][]
#     - [OpenAI](https://openai.com/)
#       - [ragna.assistants.Gpt35Turbo16k][]
#       - [ragna.assistants.Gpt4][]
#     - [AI21 Labs](https://www.ai21.com/)
#       - [ragna.assistants.Jurassic2Ultra][]
#     - [MosaicML](https://www.mosaicml.com/)
#       - [ragna.assistants.Mpt7bInstruct][]
#       - [ragna.assistants.Mpt30bInstruct][]
#
#     !!! note
#
#         To use any of builtin assistants, you need to
#         [procure API keys](../../references/faq.md#where-do-i-get-api-keys-for-the-builtin-assistants)
#         first and set the corresponding environment variables.

# %%
# ## Step 4: Start chatting
#
# We now have all parts to start a chat.

from ragna import Rag

chat = Rag().chat(
    documents=[document_path],
    source_storage=RagnaDemoSourceStorage,
    assistant=RagnaDemoAssistant,
)

# %%
# Before we can ask a question, we need to [`prepare`][ragna.core.Chat.prepare] the chat, which under the hood
# stores the documents we have selected in the source storage.

_ = await chat.prepare()

# %%
# !!! note
#
#     Ragna chats are asynchronous for better performance in real-world scenarios. You
#     can check out
#     [Python's asyncio documentation](https://docs.python.org/3/library/asyncio.html)
#     for more information. In practice, you don't need to understand all the details,
#     only use the `async` and `await` keywords with the function definition and call
#     respectively.

# %%
# Finally, we can get an [`answer`][ragna.core.Chat.answer] to a question.

print(await chat.answer("What is Ragna?"))



import itertools
from collections import defaultdict
from pathlib import Path
from types import ModuleType
from typing import Annotated, Iterable, Type, TypeVar, cast

import pydantic
import questionary
import rich
import typer
from rich.markup import escape
from rich.table import Table

import ragna
from ragna.core import (
    Assistant,
    Component,
    EnvVarRequirement,
    PackageRequirement,
    RagnaException,
    Requirement,
    SourceStorage,
)
from ragna.deploy import Config


def parse_config(value: str) -> Config:
    try:
        config = Config.from_file(value)
    except RagnaException:
        rich.print(f"The configuration file {value} does not exist.")
        if value == "./ragna.toml":
            rich.print(
                "If you don't have a configuration file yet, "
                "run [bold]ragna init[/bold] to generate one."
            )
        raise typer.Exit(1)
    except pydantic.ValidationError as validation:
        # FIXME: pretty formatting!
        for error in validation.errors():
            rich.print(error)
        raise typer.Exit(1)
    # This stores the original value so we can pass it on to subprocesses that we might
    # start.
    config.__ragna_cli_config_path__ = value
    return config


ConfigOption = Annotated[
    Config,
    typer.Option(
        "-c",
        "--config",
        metavar="CONFIG",
        envvar="RAGNA_CONFIG",
        parser=parse_config,
        help="Path to a Ragna configuration file.",
    ),
]

# This adds a newline before every question to unclutter the output
QMARK = "\n?"


def init_config(*, output_path: Path, force: bool) -> tuple[Config, Path, bool]:
    # FIXME: add link to the config documentation when it is available
    rich.print(
        "\n\t[bold]Welcome to the Ragna config creation wizard![/bold]\n\n"
        "I'll help you create a configuration file to use with ragna.\n"
        "Due to the large amount of parameters, "
        "I unfortunately can't cover everything. "
        "If you want to customize everything, "
        "please have a look at the documentation instead."
    )

    intent = questionary.select(
        "Which of the following statements describes best what you want to do?",
        choices=[
            questionary.Choice(
                (
                    "I want to try Ragna without worrying about any additional "
                    "dependencies or setup."
                ),
                value="demo",
            ),
            questionary.Choice(
                (
                    "I want to try Ragna and its builtin source storages and "
                    "assistants, which potentially require additional dependencies "
                    "or setup."
                ),
                value="builtin",
            ),
            questionary.Choice(
                (
                    "I have used Ragna before and want to customize the most common "
                    "parameters."
                ),
                value="common",
            ),
        ],
        qmark=QMARK,
    ).unsafe_ask()

    wizard = {
        "demo": _wizard_demo,
        "builtin": _wizard_builtin,
        "common": _wizard_common,
    }[intent]
    config = wizard()

    if output_path.exists() and not force:
        output_path, force = _handle_output_path(output_path=output_path, force=force)

    rich.print(
        f"\nAnd with that we are done :tada: "
        f"I'm writing the configuration file to {output_path}."
    )

    return config, output_path, force


def _wizard_demo() -> Config:
    return Config()


def _wizard_builtin() -> Config:
    config = _wizard_demo()

    rich.print(
        "\nragna has the following components builtin. "
        "Select the ones that you want to use. "
        "If the requirements of a selected component are not met, "
        "I'll show you instructions how to meet them later."
    )
    config.source_storages = _select_components(
        "source storages",
        ragna.source_storages,
        SourceStorage,  # type: ignore[type-abstract]
    )
    config.assistants = _select_components(
        "assistants",
        ragna.assistants,
        Assistant,  # type: ignore[type-abstract]
    )

    _handle_unmet_requirements(
        itertools.chain(config.source_storages, config.assistants)
    )

    return config


T = TypeVar("T", bound=Component)


def _select_components(
    title: str,
    module: ModuleType,
    base_cls: Type[T],
) -> list[Type[T]]:
    components = sorted(
        (
            obj
            for obj in module.__dict__.values()
            if isinstance(obj, type)
            and issubclass(obj, base_cls)
            and obj is not base_cls
        ),
        key=lambda component: component.display_name(),
    )
    return cast(
        list[Type[T]],
        questionary.checkbox(
            f"Which {title} do you want to use?",
            choices=[
                questionary.Choice(
                    component.display_name(),
                    value=component,
                    checked=component.is_available(),
                )
                for component in components
            ],
            qmark=QMARK,
        ).unsafe_ask(),
    )


def _handle_unmet_requirements(components: Iterable[Type[Component]]) -> None:
    unmet_requirements = set(
        requirement
        for component in components
        for requirement in component.requirements()
        if not requirement.is_available()
    )
    if not unmet_requirements:
        return

    rich.print(
        "You have selected components, which have additional requirements that are"
        "currently not met."
    )
    unmet_requirements_by_type = _split_requirements(unmet_requirements)

    unmet_package_requirements = sorted(
        str(requirement)
        for requirement in unmet_requirements_by_type[PackageRequirement]
    )
    if unmet_package_requirements:
        rich.print(
            "\nTo use the selected components, "
            "you need to install the following packages: \n"
        )
        for requirement in unmet_package_requirements:
            rich.print(f"- {requirement}")

        rich.print(
            f"\nTo do this, you can run\n\n"
            f"$ pip install {' '.join(unmet_package_requirements)}\n\n"
            f"Optionally, you can also install Ragna with all optional dependencies"
            f"for the builtin components\n\n"
            f"$ pip install '{escape('ragna[all]')}"
        )

    unmet_env_var_requirements = sorted(
        str(requirement)
        for requirement in unmet_requirements_by_type[EnvVarRequirement]
    )
    if unmet_env_var_requirements:
        rich.print(
            "\nTo use the selected components, "
            "you need to set the following environment variables: \n"
        )
        for requirement in unmet_env_var_requirements:
            rich.print(f"- {requirement}")

    rich.print(
        "\nTip: You can check the availability of the requirements with "
        "[bold]ragna check[/bold]."
    )


def _wizard_common() -> Config:
    config = _wizard_builtin()

    config.local_root = Path(
        questionary.path(
            "Where should local files be stored?",
            default=str(config.local_root),
            qmark=QMARK,
        ).unsafe_ask()
    )

    for sub_config, title in [(config.api, "REST API"), (config.ui, "web UI")]:
        sub_config.hostname = questionary.text(
            f"What hostname do you want to bind the the Ragna {title} to?",
            default=sub_config.hostname,  # type: ignore[attr-defined]
            qmark=QMARK,
        ).unsafe_ask()

        sub_config.port = int(
            questionary.text(
                f"What port do you want to bind the the Ragna {title} to?",
                default=str(sub_config.port),  # type: ignore[attr-defined]
                qmark=QMARK,
            ).unsafe_ask()
        )

    config.api.database_url = questionary.text(
        "What is the URL of the SQL database?",
        default=Config(local_root=config.local_root).api.database_url,
        qmark=QMARK,
    ).unsafe_ask()

    config.api.url = questionary.text(
        "At which URL will the Ragna REST API be served?",
        default=Config(
            api=dict(  # type: ignore[arg-type]
                hostname=config.api.hostname,
                port=config.api.port,
            )
        ).api.url,
        qmark=QMARK,
    ).unsafe_ask()

    config.api.origins = config.ui.origins = [
        questionary.text(
            "At which URL will the Ragna web UI be served?",
            default=Config(
                ui=dict(  # type: ignore[arg-type]
                    hostname=config.ui.hostname,
                    port=config.ui.port,
                )
            ).api.origins[0],
            qmark=QMARK,
        ).unsafe_ask()
    ]

    return config


def _handle_output_path(*, output_path: Path, force: bool) -> tuple[Path, bool]:
    rich.print(
        f"\nThe output path {output_path} already exists "
        f"and you didn't pass the --force flag to overwrite it. "
    )
    action = questionary.select(
        "What do you want to do?",
        choices=[
            questionary.Choice("Overwrite the existing file.", value="overwrite"),
            questionary.Choice("Select a new output path.", value="new"),
        ],
        qmark=QMARK,
    ).unsafe_ask()

    if action == "overwrite":
        force = True
    elif action == "new":
        while True:
            output_path = (
                Path(
                    questionary.path(
                        "Please provide a different output path "
                        "to write the generated config to:",
                        default=str(output_path),
                        qmark=QMARK,
                    ).unsafe_ask()
                )
                .expanduser()
                .resolve()
            )

            if not output_path.exists():
                break

            rich.print(f"The output path {output_path} already exists.")

    return output_path, force


def check_config(config: Config) -> bool:
    fully_available = True

    for title, components in [
        ("source storages", config.source_storages),
        ("assistants", config.assistants),
    ]:
        components = cast(list[Type[Component]], components)

        table = Table(
            "",
            "name",
            "environment variables",
            "packages",
            show_lines=True,
            title=title,
        )

        for component in components:
            is_available = component.is_available()
            fully_available &= is_available

            requirements = _split_requirements(component.requirements())
            table.add_row(
                _yes_or_no(is_available),
                component.display_name(),
                _format_requirements(requirements[EnvVarRequirement]),
                _format_requirements(requirements[PackageRequirement]),
            )

        rich.print(table)

    return fully_available


def _split_requirements(
    requirements: Iterable[Requirement],
) -> dict[Type[Requirement], list[Requirement]]:
    split_reqs = defaultdict(list)
    for req in requirements:
        split_reqs[type(req)].append(req)
    return split_reqs


def _format_requirements(requirements: list[Requirement]) -> str:
    if not requirements:
        return ""

    return "\n".join(f"{_yes_or_no(req.is_available())} {req}" for req in requirements)


def _yes_or_no(condition: bool) -> str:
    return ":white_check_mark:" if condition else ":x:"



import pydantic
import pytest

from ragna import Rag, assistants, source_storages
from ragna.core import LocalDocument


@pytest.fixture()
def demo_document(tmp_path, request):
    path = tmp_path / "demo_document.txt"
    with open(path, "w") as file:
        file.write(f"{request.node.name}\n")
    return LocalDocument.from_path(path)


class TestChat:
    def chat(self, documents, **params):
        return Rag().chat(
            documents=documents,
            source_storage=source_storages.RagnaDemoSourceStorage,
            assistant=assistants.RagnaDemoAssistant,
            **params,
        )

    def test_extra_params(self, demo_document):
        with pytest.raises(pydantic.ValidationError, match="not_supported_parameter"):
            self.chat(
                documents=[demo_document], not_supported_parameter="arbitrary_value"
            )

    def test_document_path(self, demo_document):
        chat = self.chat(documents=[demo_document.path])

        assert len(chat.documents) == 1
        document = chat.documents[0]

        assert isinstance(document, LocalDocument)
        assert document.path == demo_document.path
        assert document.name == demo_document.name



from __future__ import annotations

import datetime
import inspect
import uuid
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

import pydantic
from starlette.concurrency import iterate_in_threadpool, run_in_threadpool

from ._components import Assistant, Component, Message, MessageRole, SourceStorage
from ._document import Document, LocalDocument
from ._utils import RagnaException, default_user, merge_models

T = TypeVar("T")
C = TypeVar("C", bound=Component)


class Rag(Generic[C]):
    """RAG workflow.

    !!! tip

        This class can be imported from `ragna` directly, e.g.

        ```python
        from ragna import Rag
        ```
    """

    def __init__(self) -> None:
        self._components: dict[Type[C], C] = {}

    def _load_component(
        self, component: Union[Type[C], C], *, ignore_unavailable: bool = False
    ) -> Optional[C]:
        cls: Type[C]
        instance: Optional[C]

        if isinstance(component, Component):
            instance = cast(C, component)
            cls = type(instance)
        elif isinstance(component, type) and issubclass(component, Component):
            cls = component
            instance = None
        else:
            raise RagnaException

        if cls not in self._components:
            if instance is None:
                if not cls.is_available():
                    if ignore_unavailable:
                        return None

                    raise RagnaException(
                        "Component not available", name=cls.display_name()
                    )
                instance = cls()

            self._components[cls] = instance

        return self._components[cls]

    def chat(
        self,
        *,
        documents: Iterable[Any],
        source_storage: Union[Type[SourceStorage], SourceStorage],
        assistant: Union[Type[Assistant], Assistant],
        **params: Any,
    ) -> Chat:
        """Create a new [ragna.core.Chat][].

        Args:
            documents: Documents to use. If any item is not a [ragna.core.Document][],
                [ragna.core.LocalDocument.from_path][] is invoked on it.
            source_storage: Source storage to use.
            assistant: Assistant to use.
            **params: Additional parameters passed to the source storage and assistant.
        """
        return Chat(
            self,
            documents=documents,
            source_storage=source_storage,
            assistant=assistant,
            **params,
        )


class SpecialChatParams(pydantic.BaseModel):
    user: str = pydantic.Field(default_factory=default_user)
    chat_id: uuid.UUID = pydantic.Field(default_factory=uuid.uuid4)
    chat_name: str = pydantic.Field(
        default_factory=lambda: f"Chat {datetime.datetime.now():%x %X}"
    )


class Chat:
    """
    !!! tip

        This object is usually not instantiated manually, but rather through
        [ragna.core.Rag.chat][].

    A chat needs to be [`prepare`][ragna.core.Chat.prepare]d before prompts can be
    [`answer`][ragna.core.Chat.answer]ed.

    Can be used as context manager to automatically invoke
    [`prepare`][ragna.core.Chat.prepare]:

    ```python
    rag = Rag()

    async with rag.chat(
        documents=[path],
        source_storage=ragna.core.RagnaDemoSourceStorage,
        assistant=ragna.core.RagnaDemoAssistant,
    ) as chat:
        print(await chat.answer("What is Ragna?"))
    ```

    Args:
        rag: The RAG workflow this chat is associated with.
        documents: Documents to use. If any item is not a [ragna.core.Document][],
            [ragna.core.LocalDocument.from_path][] is invoked on it.
        source_storage: Source storage to use.
        assistant: Assistant to use.
        **params: Additional parameters passed to the source storage and assistant.
    """

    def __init__(
        self,
        rag: Rag,
        *,
        documents: Iterable[Any],
        source_storage: Union[Type[SourceStorage], SourceStorage],
        assistant: Union[Type[Assistant], Assistant],
        **params: Any,
    ) -> None:
        self._rag = rag

        self.documents = self._parse_documents(documents)
        self.source_storage = cast(
            SourceStorage, self._rag._load_component(source_storage)
        )
        self.assistant = cast(Assistant, self._rag._load_component(assistant))

        special_params = SpecialChatParams().model_dump()
        special_params.update(params)
        params = special_params
        self.params = params
        self._unpacked_params = self._unpack_chat_params(params)

        self._prepared = False
        self._messages: list[Message] = []

    async def prepare(self) -> Message:
        """Prepare the chat.

        This [`store`][ragna.core.SourceStorage.store]s the documents in the selected
        source storage. Afterwards prompts can be [`answer`][ragna.core.Chat.answer]ed.

        Returns:
            Welcome message.

        Raises:
            ragna.core.RagnaException: If chat is already
                [`prepare`][ragna.core.Chat.prepare]d.
        """
        if self._prepared:
            raise RagnaException(
                "Chat is already prepared",
                chat=self,
                http_status_code=400,
                detail=RagnaException.EVENT,
            )

        await self._run(self.source_storage.store, self.documents)
        self._prepared = True

        welcome = Message(
            content="How can I help you with the documents?",
            role=MessageRole.SYSTEM,
        )
        self._messages.append(welcome)
        return welcome

    async def answer(self, prompt: str, *, stream: bool = False) -> Message:
        """Answer a prompt.

        Returns:
            Answer.

        Raises:
            ragna.core.RagnaException: If chat is not
                [`prepare`][ragna.core.Chat.prepare]d.
        """
        if not self._prepared:
            raise RagnaException(
                "Chat is not prepared",
                chat=self,
                http_status_code=400,
                detail=RagnaException.EVENT,
            )

        self._messages.append(Message(content=prompt, role=MessageRole.USER))

        sources = await self._run(self.source_storage.retrieve, self.documents, prompt)

        answer = Message(
            content=self._run_gen(self.assistant.answer, prompt, sources),
            role=MessageRole.ASSISTANT,
            sources=sources,
        )
        if not stream:
            await answer.read()

        self._messages.append(answer)

        return answer

    def _parse_documents(self, documents: Iterable[Any]) -> list[Document]:
        documents_ = []
        for document in documents:
            if not isinstance(document, Document):
                document = LocalDocument.from_path(document)

            if not document.is_readable():
                raise RagnaException(
                    "Document not readable",
                    document=document,
                    http_status_code=404,
                )

            documents_.append(document)
        return documents_

    def _unpack_chat_params(
        self, params: dict[str, Any]
    ) -> dict[Callable, dict[str, Any]]:
        component_models = {
            getattr(component, name): model
            for component in [self.source_storage, self.assistant]
            for (_, name), model in component._protocol_models().items()
        }

        ChatModel = merge_models(
            str(self.params["chat_id"]),
            SpecialChatParams,
            *component_models.values(),
            config=pydantic.ConfigDict(extra="forbid"),
        )

        chat_params = ChatModel.model_validate(params, strict=True).model_dump(
            exclude_none=True
        )
        return {
            fn: model(**chat_params).model_dump()
            for fn, model in component_models.items()
        }

    async def _run(
        self, fn: Union[Callable[..., T], Callable[..., Awaitable[T]]], *args: Any
    ) -> T:
        kwargs = self._unpacked_params[fn]
        if inspect.iscoroutinefunction(fn):
            fn = cast(Callable[..., Awaitable[T]], fn)
            coro = fn(*args, **kwargs)
        else:
            fn = cast(Callable[..., T], fn)
            coro = run_in_threadpool(fn, *args, **kwargs)

        return await coro

    async def _run_gen(
        self,
        fn: Union[Callable[..., Iterator[T]], Callable[..., AsyncIterator[T]]],
        *args: Any,
    ) -> AsyncIterator[T]:
        kwargs = self._unpacked_params[fn]
        if inspect.isasyncgenfunction(fn):
            fn = cast(Callable[..., AsyncIterator[T]], fn)
            async_gen = fn(*args, **kwargs)
        else:
            fn = cast(Callable[..., Iterator[T]], fn)
            async_gen = iterate_in_threadpool(fn(*args, **kwargs))

        async for item in async_gen:
            yield item

    async def __aenter__(self) -> Chat:
        await self.prepare()
        return self

    async def __aexit__(
        self, exc_type: Type[Exception], exc: Exception, traceback: str
    ) -> None:
        pass



import pytest
from fastapi import status
from fastapi.testclient import TestClient

from ragna import assistants
from ragna.core import RagnaException
from ragna.deploy import Config
from ragna.deploy._api import app

from .utils import authenticate


@pytest.mark.parametrize("ignore_unavailable_components", [True, False])
def test_ignore_unavailable_components(ignore_unavailable_components):
    available_assistant = assistants.RagnaDemoAssistant
    assert available_assistant.is_available()

    unavailable_assistant = assistants.Gpt4
    assert not unavailable_assistant.is_available()

    config = Config(assistants=[available_assistant, unavailable_assistant])

    if ignore_unavailable_components:
        with TestClient(
            app(
                config=config,
                ignore_unavailable_components=ignore_unavailable_components,
            )
        ) as client:
            authenticate(client)

            components = client.get("/components").raise_for_status().json()
            assert [assistant["title"] for assistant in components["assistants"]] == [
                available_assistant.display_name()
            ]
    else:
        with pytest.raises(RagnaException, match="not available"):
            app(
                config=config,
                ignore_unavailable_components=ignore_unavailable_components,
            )


def test_ignore_unavailable_components_at_least_one():
    unavailable_assistant = assistants.Gpt4
    assert not unavailable_assistant.is_available()

    config = Config(assistants=[unavailable_assistant])

    with pytest.raises(RagnaException, match="No component available"):
        app(
            config=config,
            ignore_unavailable_components=True,
        )


def test_unknown_component(tmp_local_root):
    config = Config(local_root=tmp_local_root)

    document_root = config.local_root / "documents"
    document_root.mkdir()
    document_path = document_root / "test.txt"
    with open(document_path, "w") as file:
        file.write("!\n")

    with TestClient(
        app(config=Config(), ignore_unavailable_components=False)
    ) as client:
        authenticate(client)

        document_upload = (
            client.post("/document", json={"name": document_path.name})
            .raise_for_status()
            .json()
        )
        document = document_upload["document"]
        assert document["name"] == document_path.name

        parameters = document_upload["parameters"]
        with open(document_path, "rb") as file:
            client.request(
                parameters["method"],
                parameters["url"],
                data=parameters["data"],
                files={"file": file},
            )

        response = client.post(
            "/chats",
            json={
                "name": "test-chat",
                "source_storage": "unknown_source_storage",
                "assistant": "unknown_assistant",
                "params": {},
                "documents": [document],
            },
        )

        assert response.status_code == status.HTTP_404_NOT_FOUND

        error = response.json()["error"]
        assert "Unknown component" in error["message"]



import json
from typing import AsyncIterator, cast

from ragna.core import PackageRequirement, RagnaException, Requirement, Source

from ._api import ApiAssistant


class AnthropicApiAssistant(ApiAssistant):
    _API_KEY_ENV_VAR = "ANTHROPIC_API_KEY"
    _MODEL: str

    @classmethod
    def _extra_requirements(cls) -> list[Requirement]:
        return [PackageRequirement("httpx_sse")]

    @classmethod
    def display_name(cls) -> str:
        return f"Anthropic/{cls._MODEL}"

    def _instructize_prompt(self, prompt: str, sources: list[Source]) -> str:
        # See https://docs.anthropic.com/claude/docs/introduction-to-prompt-design#human--assistant-formatting
        instruction = (
            "\n\nHuman: "
            "Use the following pieces of context to answer the question at the end. "
            "If you don't know the answer, just say so. Don't try to make up an answer.\n"
        )
        instruction += "\n\n".join(source.content for source in sources)
        return f"{instruction}\n\nQuestion: {prompt}\n\nAssistant:"

    async def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> AsyncIterator[str]:
        import httpx_sse

        # See https://docs.anthropic.com/claude/reference/streaming
        async with httpx_sse.aconnect_sse(
            self._client,
            "POST",
            "https://api.anthropic.com/v1/complete",
            headers={
                "accept": "application/json",
                "anthropic-version": "2023-06-01",
                "content-type": "application/json",
                "x-api-key": self._api_key,
            },
            json={
                "model": self._MODEL,
                "prompt": self._instructize_prompt(prompt, sources),
                "max_tokens_to_sample": max_new_tokens,
                "temperature": 0.0,
                "stream": True,
            },
        ) as event_source:
            await self._assert_api_call_is_success(event_source.response)

            async for sse in event_source.aiter_sse():
                data = json.loads(sse.data)
                if data["type"] != "completion":
                    continue
                elif "error" in data:
                    raise RagnaException(data["error"].pop("message"), **data["error"])
                elif data["stop_reason"] is not None:
                    break

                yield cast(str, data["completion"])


class ClaudeInstant(AnthropicApiAssistant):
    """[Claude Instant](https://docs.anthropic.com/claude/reference/selecting-a-model)

    !!! info "Required environment variables"

        - `ANTHROPIC_API_KEY`

    !!! info "Required packages"

        - `httpx_sse`
    """

    _MODEL = "claude-instant-1"


class Claude(AnthropicApiAssistant):
    """[Claude](https://docs.anthropic.com/claude/reference/selecting-a-model)

    !!! info "Required environment variables"

        - `ANTHROPIC_API_KEY`

    !!! info "Required packages"

        - `httpx_sse`
    """

    _MODEL = "claude-2"



__all__ = [
    "Authentication",
    "Config",
    "RagnaDemoAuthentication",
]

from ._authentication import Authentication, RagnaDemoAuthentication
from ._config import Config

# isort: split

from ragna._utils import fix_module

fix_module(globals())
del fix_module



import uuid

import ragna
from ragna.core import Document, PackageRequirement, Requirement, Source

from ._vector_database import VectorDatabaseSourceStorage


class LanceDB(VectorDatabaseSourceStorage):
    """[LanceDB vector database](https://lancedb.com/)

    !!! info "Required packages"

        - `chromadb>=0.4.13`
        - `lancedb>=0.2`
        - `pyarrow`
    """

    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [
            *super().requirements(),
            PackageRequirement("lancedb>=0.2"),
            PackageRequirement(
                "pyarrow",
                # See https://github.com/apache/arrow/issues/38167
                exclude_modules=["__dummy__"],
            ),
        ]

    def __init__(self) -> None:
        super().__init__()

        import lancedb
        import pyarrow as pa

        self._db = lancedb.connect(ragna.local_root() / "lancedb")
        self._schema = pa.schema(
            [
                pa.field("id", pa.string()),
                pa.field("document_id", pa.string()),
                pa.field("page_numbers", pa.string()),
                pa.field("text", pa.string()),
                pa.field(
                    self._VECTOR_COLUMN_NAME,
                    pa.list_(pa.float32(), self._embedding_dimensions),
                ),
                pa.field("num_tokens", pa.int32()),
            ]
        )

    _VECTOR_COLUMN_NAME = "embedded_text"

    def store(
        self,
        documents: list[Document],
        *,
        chat_id: uuid.UUID,
        chunk_size: int = 500,
        chunk_overlap: int = 250,
    ) -> None:
        table = self._db.create_table(name=str(chat_id), schema=self._schema)

        for document in documents:
            for chunk in self._chunk_pages(
                document.extract_pages(),
                chunk_size=chunk_size,
                chunk_overlap=chunk_overlap,
            ):
                table.add(
                    [
                        {
                            "id": str(uuid.uuid4()),
                            "document_id": str(document.id),
                            "page_numbers": self._page_numbers_to_str(
                                chunk.page_numbers
                            ),
                            "text": chunk.text,
                            self._VECTOR_COLUMN_NAME: self._embedding_function(
                                [chunk.text]
                            )[0],
                            "num_tokens": chunk.num_tokens,
                        }
                    ]
                )

    def retrieve(
        self,
        documents: list[Document],
        prompt: str,
        *,
        chat_id: uuid.UUID,
        chunk_size: int = 500,
        num_tokens: int = 1024,
    ) -> list[Source]:
        table = self._db.open_table(str(chat_id))

        # We cannot retrieve source by a maximum number of tokens. Thus, we estimate how
        # many sources we have to query. We overestimate by a factor of two to avoid
        # retrieving to few sources and needed to query again.
        limit = int(num_tokens * 2 / chunk_size)
        results = (
            table.search(
                self._embedding_function([prompt])[0],
                vector_column_name=self._VECTOR_COLUMN_NAME,
            )
            .limit(limit)
            .to_arrow()
        )

        document_map = {str(document.id): document for document in documents}
        return self._take_sources_up_to_max_tokens(
            (
                Source(
                    id=result["id"],
                    document=document_map[result["document_id"]],
                    # For some reason adding an empty string during store() results
                    # in this field being None. Thus, we need to parse it back here.
                    # TODO: See if there is a configuration option for this
                    location=result["page_numbers"] or "",
                    content=result["text"],
                    num_tokens=result["num_tokens"],
                )
                for result in results.to_pylist()
            ),
            max_tokens=num_tokens,
        )



from sqlalchemy import Column, ForeignKey, Table, types
from sqlalchemy.orm import DeclarativeBase, relationship  # type: ignore[attr-defined]

from ragna.core import MessageRole


class Base(DeclarativeBase):
    pass


# FIXME: Do we actually need this table? If we are sure that a user is unique and has to
#  be authenticated from the API layer, it seems having an extra mapping here is not
#  needed?
class User(Base):
    __tablename__ = "users"

    id = Column(types.Uuid, primary_key=True)  # type: ignore[attr-defined]
    name = Column(types.String)


document_chat_association_table = Table(
    "document_chat_association_table",
    Base.metadata,
    Column("document_id", ForeignKey("documents.id"), primary_key=True),
    Column("chat_id", ForeignKey("chats.id"), primary_key=True),
)


class Document(Base):
    __tablename__ = "documents"

    id = Column(types.Uuid, primary_key=True)  # type: ignore[attr-defined]
    user_id = Column(ForeignKey("users.id"))
    name = Column(types.String)
    # Mind the trailing underscore here. Unfortunately, this is necessary, because
    # metadata without the underscore is reserved by SQLAlchemy
    metadata_ = Column(types.JSON)
    chats = relationship(
        "Chat",
        secondary=document_chat_association_table,
        back_populates="documents",
    )
    sources = relationship(
        "Source",
        back_populates="document",
    )


class Chat(Base):
    __tablename__ = "chats"

    id = Column(types.Uuid, primary_key=True)  # type: ignore[attr-defined]
    user_id = Column(ForeignKey("users.id"))
    name = Column(types.String)
    documents = relationship(
        "Document",
        secondary=document_chat_association_table,
        back_populates="chats",
    )
    source_storage = Column(types.String)
    assistant = Column(types.String)
    params = Column(types.JSON)
    messages = relationship("Message", cascade="all, delete")
    prepared = Column(types.Boolean)


source_message_association_table = Table(
    "source_message_association_table",
    Base.metadata,
    Column("source_id", ForeignKey("sources.id"), primary_key=True),
    Column("message_id", ForeignKey("messages.id"), primary_key=True),
)


class Source(Base):
    __tablename__ = "sources"

    # This is not a UUID column as all of the other IDs, because we don't control its
    # generation. It is generated as part of ragna.core.SourceStorage.retrieve and using
    # a string here doesn't impose unnecessary constraints on the user.
    id = Column(types.String, primary_key=True)

    document_id = Column(ForeignKey("documents.id"))
    document = relationship("Document", back_populates="sources")

    location = Column(types.String)
    content = Column(types.String)
    num_tokens = Column(types.Integer)

    messages = relationship(
        "Message",
        secondary=source_message_association_table,
        back_populates="sources",
    )


class Message(Base):
    __tablename__ = "messages"

    id = Column(types.Uuid, primary_key=True)  # type: ignore[attr-defined]
    chat_id = Column(ForeignKey("chats.id"))
    content = Column(types.String)
    role = Column(types.Enum(MessageRole))
    sources = relationship(
        "Source",
        secondary=source_message_association_table,
        back_populates="messages",
    )
    timestamp = Column(types.DateTime)



try:
    from ._version import __version__
except ModuleNotFoundError:
    import warnings

    warnings.warn("ragna was not properly installed!")
    del warnings

    __version__ = "UNKNOWN"

from ._utils import local_root

# isort: split

from . import assistants, core, deploy, source_storages
from .core import Rag

__all__ = [
    "__version__",
    "Rag",
    "assistants",
    "core",
    "deploy",
    "local_root",
    "source_storages",
]



import uuid

import pytest

from ragna.core import LocalDocument
from ragna.source_storages import Chroma, LanceDB


@pytest.mark.parametrize("source_storage_cls", [Chroma, LanceDB])
def test_smoke(tmp_local_root, source_storage_cls):
    document_root = tmp_local_root / "documents"
    document_root.mkdir()
    documents = []
    for idx in range(10):
        path = document_root / f"irrelevant{idx}.txt"
        with open(path, "w") as file:
            file.write(f"This is irrelevant information for the {idx}. time!\n")

        documents.append(LocalDocument.from_path(path))

    secret = "Ragna"
    path = document_root / "secret.txt"
    with open(path, "w") as file:
        file.write(f"The secret is {secret}!\n")

    documents.insert(len(documents) // 2, LocalDocument.from_path(path))

    source_storage = source_storage_cls()

    # Hardcoding a chat_id here only works because all tested source storages only
    # require this.
    # TODO: make this more flexible by taking required parameters as part of the
    #  parametrization.
    chat_id = uuid.uuid4()

    source_storage.store(documents, chat_id=chat_id)

    prompt = "What is the secret?"
    sources = source_storage.retrieve(documents, prompt, chat_id=chat_id)

    assert secret in sources[0].content



from textwrap import dedent

from ragna.deploy._ui.js_utils import preformat, redirect_script


def test_preformat_basic():
    output = preformat("{ This is awesome {{var}} }")
    assert output == "{{ This is awesome {var} }}"


def test_preformat_basic_fmt():
    output = preformat("{ This is awesome {{var}} }").format(var="test")
    assert output == "{ This is awesome test }"


def test_preformat_multivars():
    output = preformat("{ {{var1}} This is awesome {{var2}} }").format(
        var1="test1", var2="test2"
    )
    assert output == "{ test1 This is awesome test2 }"


def test_preformat_unsubs():
    output = preformat("{ This is {Hello} awesome {{var}} }").format(var="test")
    assert output == "{ This is {Hello} awesome test }"


def test_redirect_script():
    output = redirect_script(remove="foo", append="bar")
    expected = dedent(
        r"""
        <script>
            var currentPath = window.location.pathname; // Get the current path
            if (currentPath.includes('/foo')) {
              if ("foo") {
                currentPath = currentPath.replace(/\/foo(\/)?$/, '')
              }
              var redirectTo = currentPath + 'bar';
              window.location.href = redirectTo;
              if (false) {
                document.cookie = "auth_token=; expires=Thu, 01 Jan 1970 00:00:00 UTC; path=${redirectTo}";
              }
            }
        </script>
    """
    )
    assert dedent(output.object) == expected



import panel as pn
import param

from ragna.deploy._ui.js_utils import redirect_script


class LogoutPage(pn.viewable.Viewer, param.Parameterized):
    def __init__(self, api_wrapper, **params):
        super().__init__(**params)
        self.api_wrapper = api_wrapper

        self.api_wrapper.auth_token = None

    def __panel__(self):
        # Usually, we do a redirect this way :
        # >>> pn.state.location.param.update(reload=True, pathname="/")
        # But it only works once the page is fully loaded.
        # So we render a javascript redirect instead.

        # To remove the token from the cookie, we have to force its expiry date to the past.
        return redirect_script(remove="logout", append="/", remove_auth_cookie=True)



"""
# Web UI

Ragna comes with a web UI where you can try out all of the features!

!!! tip "Under development"

    This documentation page is being actively worked on.
    It will be updated with more details soon. :)
"""

# %%
# You can launch the web application from the CLI:
#
# ```bash
# ragna ui
# ```
#
# It will open in a browser window automatically and you should see the login screen:
#
# ![ragna UI login screen with a username and password form. Both the username and
# password are filled with the string "ragna" and the password is visible and not
# censored](../../assets/images/web-ui-login.png)
#
# By default [ragna.deploy.RagnaDemoAuthentication][] is used. Thus, you can use any
# username and a matching password to authenticate. For example, you can leave both fields
# blank or use `ragna` / `ragna` as in the image above.



__all__ = [
    "Chroma",
    "LanceDB",
    "RagnaDemoSourceStorage",
]

from ._chroma import Chroma
from ._demo import RagnaDemoSourceStorage
from ._lancedb import LanceDB

# isort: split

from ragna._utils import fix_module

fix_module(globals())
del fix_module



from __future__ import annotations

import functools
import uuid
from typing import Any, Callable, Optional, cast
from urllib.parse import urlsplit

from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker as _sessionmaker

from ragna.core import RagnaException

from . import orm, schemas


def get_sessionmaker(database_url: str) -> Callable[[], Session]:
    components = urlsplit(database_url)
    if components.scheme == "sqlite":
        connect_args = dict(check_same_thread=False)
    else:
        connect_args = dict()
    engine = create_engine(database_url, connect_args=connect_args)
    orm.Base.metadata.create_all(bind=engine)
    return _sessionmaker(autocommit=False, autoflush=False, bind=engine)


@functools.lru_cache(maxsize=1024)
def _get_user_id(session: Session, username: str) -> uuid.UUID:
    user: Optional[orm.User] = session.execute(
        select(orm.User).where(orm.User.name == username)
    ).scalar_one_or_none()

    if user is None:
        # Add a new user if the current username is not registered yet. Since this is
        # behind the authentication layer, we don't need any extra security here.
        user = orm.User(id=uuid.uuid4(), name=username)
        session.add(user)
        session.commit()

    return cast(uuid.UUID, user.id)


def add_document(
    session: Session, *, user: str, document: schemas.Document, metadata: dict[str, Any]
) -> None:
    session.add(
        orm.Document(
            id=document.id,
            user_id=_get_user_id(session, user),
            name=document.name,
            metadata_=metadata,
        )
    )
    session.commit()


def _orm_to_schema_document(document: orm.Document) -> schemas.Document:
    return schemas.Document(id=document.id, name=document.name)


@functools.lru_cache(maxsize=1024)
def get_document(
    session: Session, *, user: str, id: uuid.UUID
) -> tuple[schemas.Document, dict[str, Any]]:
    document = session.execute(
        select(orm.Document).where(
            (orm.Document.user_id == _get_user_id(session, user))
            & (orm.Document.id == id)
        )
    ).scalar_one_or_none()
    return _orm_to_schema_document(document), document.metadata_


def add_chat(session: Session, *, user: str, chat: schemas.Chat) -> None:
    document_ids = {document.id for document in chat.metadata.documents}
    documents = (
        session.execute(select(orm.Document).where(orm.Document.id.in_(document_ids)))
        .scalars()
        .all()
    )
    if len(documents) != len(document_ids):
        raise RagnaException(
            str(set(document_ids) - {document.id for document in documents})
        )
    session.add(
        orm.Chat(
            id=chat.id,
            user_id=_get_user_id(session, user),
            name=chat.metadata.name,
            documents=documents,
            source_storage=chat.metadata.source_storage,
            assistant=chat.metadata.assistant,
            params=chat.metadata.params,
            prepared=chat.prepared,
        )
    )
    session.commit()


def _orm_to_schema_chat(chat: orm.Chat) -> schemas.Chat:
    documents = [
        schemas.Document(id=document.id, name=document.name)
        for document in chat.documents
    ]
    messages = [
        schemas.Message(
            id=message.id,
            role=message.role,
            content=message.content,
            sources=[
                schemas.Source(
                    id=source.id,
                    document=_orm_to_schema_document(source.document),
                    location=source.location,
                    content=source.content,
                    num_tokens=source.num_tokens,
                )
                for source in message.sources
            ],
            timestamp=message.timestamp,
        )
        for message in chat.messages
    ]
    return schemas.Chat(
        id=chat.id,
        metadata=schemas.ChatMetadata(
            name=chat.name,
            documents=documents,
            source_storage=chat.source_storage,
            assistant=chat.assistant,
            params=chat.params,  # type: ignore[arg-type]
        ),
        messages=messages,
        prepared=chat.prepared,
    )


def get_chats(session: Session, *, user: str) -> list[schemas.Chat]:
    return [
        _orm_to_schema_chat(chat)
        for chat in session.execute(
            select(orm.Chat).where(orm.Chat.user_id == _get_user_id(session, user))
        )
        .scalars()
        .all()
    ]


def _get_orm_chat(session: Session, *, user: str, id: uuid.UUID) -> orm.Chat:
    chat: Optional[orm.Chat] = session.execute(
        select(orm.Chat).where(
            (orm.Chat.id == id) & (orm.Chat.user_id == _get_user_id(session, user))
        )
    ).scalar_one_or_none()
    if chat is None:
        raise RagnaException()
    return chat


def get_chat(session: Session, *, user: str, id: uuid.UUID) -> schemas.Chat:
    return _orm_to_schema_chat(_get_orm_chat(session, user=user, id=id))


def _schema_to_orm_source(session: Session, source: schemas.Source) -> orm.Source:
    orm_source: Optional[orm.Source] = session.execute(
        select(orm.Source).where(orm.Source.id == source.id)
    ).scalar_one_or_none()

    if orm_source is None:
        orm_source = orm.Source(
            id=source.id,
            document_id=source.document.id,
            location=source.location,
            content=source.content,
            num_tokens=source.num_tokens,
        )
        session.add(orm_source)
        session.commit()
        session.refresh(orm_source)

    return orm_source


def _schema_to_orm_message(
    session: Session, chat_id: uuid.UUID, message: schemas.Message
) -> orm.Message:
    orm_message: Optional[orm.Message] = session.execute(
        select(orm.Message).where(orm.Message.id == message.id)
    ).scalar_one_or_none()
    if orm_message is None:
        orm_message = orm.Message(
            id=message.id,
            chat_id=chat_id,
            content=message.content,
            role=message.role,
            sources=[
                _schema_to_orm_source(session, source=source)
                for source in message.sources
            ],
            timestamp=message.timestamp,
        )
        session.add(orm_message)
        session.commit()
        session.refresh(orm_message)

    return orm_message


def update_chat(session: Session, user: str, chat: schemas.Chat) -> None:
    orm_chat = _get_orm_chat(session, user=user, id=chat.id)

    orm_chat.prepared = chat.prepared
    orm_chat.messages = [
        _schema_to_orm_message(session, chat_id=chat.id, message=message)
        for message in chat.messages
    ]

    session.commit()


def delete_chat(session: Session, user: str, id: uuid.UUID) -> None:
    orm_chat = _get_orm_chat(session, user=user, id=id)
    session.delete(orm_chat)  # type: ignore[no-untyped-call]
    session.commit()



from pathlib import Path
from urllib.parse import urlsplit

import panel as pn
import param

from ragna._utils import handle_localhost_origins
from ragna.deploy import Config

from . import js
from . import styles as ui
from .api_wrapper import ApiWrapper, RagnaAuthTokenExpiredException
from .auth_page import AuthPage
from .js_utils import redirect_script
from .logout_page import LogoutPage
from .main_page import MainPage

pn.extension(
    loading_spinner="dots",
    loading_color=ui.MAIN_COLOR,
    layout_compatibility="error",
)
pn.config.browser_info = True


HERE = Path(__file__).parent
# CSS = HERE / "css"
IMGS = HERE / "imgs"
RES = HERE / "resources"


class App(param.Parameterized):
    def __init__(self, *, hostname, port, api_url, origins, open_browser):
        super().__init__()
        ui.apply_design_modifiers()
        self.hostname = hostname
        self.port = port
        self.api_url = api_url
        self.origins = origins
        self.open_browser = open_browser

    def get_template(self):
        template = pn.template.FastListTemplate(
            # We need to set a title to have it appearing on the browser's tab
            # but it means we need to hide it from the header bar
            title="Ragna",
            accent_base_color=ui.MAIN_COLOR,
            theme_toggle=False,
            collapsed_sidebar=True,
            # main_layout=None
            raw_css=[ui.APP_RAW],
            favicon="imgs/ragna_logo.svg",
            css_files=["https://rsms.me/inter/inter.css"],
        )

        template.modal.objects = [
            pn.Column(
                sizing_mode="stretch_both",
            )
        ]

        template.header.append(pn.pane.HTML(js.SHADOWROOT_INDEXING))
        template.header.append(pn.pane.HTML(js.MODAL_MOUSE_UP_FIX))
        template.header.append(pn.pane.HTML(js.CONNECTION_MONITOR))

        return template

    def index_page(self):
        if "auth_token" not in pn.state.cookies:
            return redirect_script(remove="", append="auth")

        try:
            api_wrapper = ApiWrapper(
                api_url=self.api_url, auth_token=pn.state.cookies["auth_token"]
            )
        except RagnaAuthTokenExpiredException:
            # If the token has expired / is invalid, we redirect to the logout page.
            # The logout page will delete the cookie and redirect to the auth page.
            return redirect_script(remove="", append="logout")

        template = self.get_template()
        main_page = MainPage(api_wrapper=api_wrapper, template=template)
        template.main.append(main_page)
        return template

    def auth_page(self):
        # If the user is already authenticated, we receive the auth token in the cookie.
        # in that case, redirect to the index page.
        if "auth_token" in pn.state.cookies:
            # Usually, we do a redirect this way :
            # >>> pn.state.location.param.update(reload=True, pathname="/")
            # But it only works once the page is fully loaded.
            # So we render a javascript redirect instead.
            return redirect_script(remove="auth")

        template = self.get_template()
        auth_page = AuthPage(api_wrapper=ApiWrapper(api_url=self.api_url))
        template.main.append(auth_page)
        return template

    def logout_page(self):
        template = self.get_template()
        logout_page = LogoutPage(api_wrapper=ApiWrapper(api_url=self.api_url))
        template.main.append(logout_page)
        return template

    def health_page(self):
        return pn.pane.HTML("<h1>Ok</h1>")

    def serve(self):
        all_pages = {
            "/": self.index_page,
            "/auth": self.auth_page,
            "/logout": self.logout_page,
            "/health": self.health_page,
        }
        titles = {"/": "Home"}

        pn.serve(
            all_pages,
            titles=titles,
            address=self.hostname,
            port=self.port,
            admin=True,
            start=True,
            location=True,
            show=self.open_browser,
            keep_alive=30 * 1000,  # 30s
            autoreload=True,
            profiler="pyinstrument",
            allow_websocket_origin=[urlsplit(origin).netloc for origin in self.origins],
            static_dirs={"imgs": str(IMGS), "resources": str(RES)},  # "css": str(CSS),
        )


def app(*, config: Config, open_browser: bool) -> App:
    return App(
        hostname=config.ui.hostname,
        port=config.ui.port,
        api_url=config.api.url,
        origins=handle_localhost_origins(config.ui.origins),
        open_browser=open_browser,
    )



import subprocess
import sys
import time
from pathlib import Path
from typing import Annotated, Optional

import httpx
import rich
import typer
import uvicorn

import ragna
from ragna._utils import timeout_after
from ragna.deploy._api import app as api_app
from ragna.deploy._ui import app as ui_app

from .config import ConfigOption, check_config, init_config

app = typer.Typer(
    name="ragna",
    invoke_without_command=True,
    no_args_is_help=True,
    add_completion=False,
    pretty_exceptions_enable=False,
)


def version_callback(value: bool) -> None:
    if value:
        rich.print(f"ragna {ragna.__version__} from {ragna.__path__[0]}")
        raise typer.Exit()


@app.callback()
def _main(
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version", callback=version_callback, help="Show version and exit."
        ),
    ] = None,
) -> None:
    pass


@app.command(help="Start a wizard to build a Ragna configuration interactively.")
def init(
    *,
    output_path: Annotated[
        Path,
        typer.Option(
            "-o",
            "--output-file",
            metavar="OUTPUT_PATH",
            default_factory=lambda: Path.cwd() / "ragna.toml",
            show_default="./ragna.toml",
            help="Write configuration to <OUTPUT_PATH>.",
        ),
    ],
    force: Annotated[
        bool,
        typer.Option(
            "-f", "--force", help="Overwrite an existing file at <OUTPUT_PATH>."
        ),
    ] = False,
) -> None:
    config, output_path, force = init_config(output_path=output_path, force=force)
    config.to_file(output_path, force=force)


@app.command(help="Check the availability of components.")
def check(config: ConfigOption = "./ragna.toml") -> None:  # type: ignore[assignment]
    is_available = check_config(config)
    raise typer.Exit(int(not is_available))


@app.command(help="Start the REST API.")
def api(
    *,
    config: ConfigOption = "./ragna.toml",  # type: ignore[assignment]
    ignore_unavailable_components: Annotated[
        bool,
        typer.Option(
            help=(
                "Ignore components that are not available, "
                "i.e. their requirements are not met. "
            )
        ),
    ] = False,
) -> None:
    uvicorn.run(
        api_app(
            config=config, ignore_unavailable_components=ignore_unavailable_components
        ),
        host=config.api.hostname,
        port=config.api.port,
    )


@app.command(help="Start the web UI.")
def ui(
    *,
    config: ConfigOption = "./ragna.toml",  # type: ignore[assignment]
    start_api: Annotated[
        Optional[bool],
        typer.Option(
            help="Start the ragna REST API alongside the web UI in a subprocess.",
            show_default="Start if the API is not served at the configured URL.",
        ),
    ] = None,
    ignore_unavailable_components: Annotated[
        bool,
        typer.Option(
            help=(
                "Ignore components that are not available, "
                "i.e. their requirements are not met. "
                "This option as no effect if --no-start-api is used."
            )
        ),
    ] = False,
    open_browser: Annotated[
        bool,
        typer.Option(help="Open the web UI in the browser when it is started."),
    ] = True,
) -> None:
    def check_api_available() -> bool:
        try:
            return httpx.get(config.api.url).is_success
        except httpx.ConnectError:
            return False

    if start_api is None:
        start_api = not check_api_available()

    if start_api:
        process = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "ragna",
                "api",
                "--config",
                config.__ragna_cli_config_path__,  # type: ignore[attr-defined]
                f"--{'' if ignore_unavailable_components else 'no-'}ignore-unavailable-components",
            ],
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
    else:
        process = None

    try:
        if process is not None:

            @timeout_after(60)
            def wait_for_api() -> None:
                while not check_api_available():
                    time.sleep(0.5)

            try:
                wait_for_api()
            except TimeoutError:
                rich.print(
                    "Failed to start the API in 60 seconds. "
                    "Please start it manually with [bold]ragna api[/bold]."
                )
                raise typer.Exit(1)

        ui_app(config=config, open_browser=open_browser).serve()  # type: ignore[no-untyped-call]
    finally:
        if process is not None:
            process.kill()
            process.communicate()



__all__ = [
    "Assistant",
    "Chat",
    "Component",
    "Document",
    "DocumentHandler",
    "DocumentUploadParameters",
    "DocxDocumentHandler",
    "PptxDocumentHandler",
    "EnvVarRequirement",
    "LocalDocument",
    "Message",
    "MessageRole",
    "PackageRequirement",
    "Page",
    "PdfDocumentHandler",
    "Rag",
    "RagnaException",
    "Requirement",
    "Source",
    "SourceStorage",
    "PlainTextDocumentHandler",
]

from ._utils import (
    EnvVarRequirement,
    PackageRequirement,
    RagnaException,
    Requirement,
)

# isort: split

from ._document import (
    Document,
    DocumentHandler,
    DocumentUploadParameters,
    DocxDocumentHandler,
    LocalDocument,
    Page,
    PdfDocumentHandler,
    PlainTextDocumentHandler,
    PptxDocumentHandler,
)

# isort: split

from ._components import (
    Assistant,
    Component,
    Message,
    MessageRole,
    Source,
    SourceStorage,
)

# isort: split

from ._rag import Chat, Rag

# isort: split

from ragna._utils import fix_module

fix_module(globals())
del fix_module



import textwrap
import uuid

from ragna.core import Document, Source, SourceStorage


class RagnaDemoSourceStorage(SourceStorage):
    """Demo assistant without requirements.

    !!! note

        As the name implies, this source storage is just for demo purposes and cannot
        retrieve relevant sources for a given prompt. It returns a single
        [ragna.core.Source][] per stored [ragna.core.Document][] with potentially
        shortened text extracted from the first [ragna.core.Page][].
    """

    @classmethod
    def display_name(cls) -> str:
        return "Ragna/DemoSourceStorage"

    def __init__(self) -> None:
        self._storage: dict[uuid.UUID, list[Source]] = {}

    def store(self, documents: list[Document], *, chat_id: uuid.UUID) -> None:
        self._storage[chat_id] = [
            Source(
                id=str(uuid.uuid4()),
                document=document,
                location=f"page {page.number}"
                if (page := next(document.extract_pages())).number
                else "",
                content=(content := textwrap.shorten(page.text, width=100)),
                num_tokens=len(content.split()),
            )
            for document in documents
        ]

    def retrieve(
        self, documents: list[Document], prompt: str, *, chat_id: uuid.UUID
    ) -> list[Source]:
        return self._storage[chat_id]



"""
# REST API

Ragna was designed to help you quickly build custom RAG powered web
applications. For this you can leverage the built-in
[REST API](../../references/rest-api.md).

This tutorial walks you through basic steps of using Ragnas REST API.
"""

# %%
# Before we start this tutorial, we import some helpers.

import sys
from pathlib import Path

sys.path.insert(0, str(Path.cwd().parent))

import documentation_helpers

# %%
# ## Step 1: Start the REST API
#
# Ragnas REST API is normally started from a terminal with
#
# ```bash
# $ ragna api
# ```
#
# For this tutorial we use our helper that does the equivalent just from Python.
#
# !!! note
#
#     By default, the REST API is started from the `ragna.toml` configuration file in
#     the current working directory. If you don't have a configuration file yet, you can
#     run
#
#     ```bash
#     $ ragna init
#     ```
#
#     to start an interactive wizard that helps you create one. The config that we'll
#     be using for this tutorial is equivalent of picking the first option the wizard
#     offers you, i.e. using only demo components.

from ragna.deploy import Config

config = Config()

rest_api = documentation_helpers.RestApi()
_ = rest_api.start(config)

# %%
# Let's make sure the REST API is started correctly and can be reached.

import httpx

client = httpx.Client(base_url=config.api.url)
client.get("/").raise_for_status()


# %%
# ## Step 2: Authentication
#
# In order to use Ragnas REST API, we need to authenticate first. To forge an API token
# we send a request to the `/token` endpoint. This is processed by the
# [`Authentication`][ragna.deploy.Authentication], which can be overridden through the
# config. For this tutorial, we use the default
# [ragna.deploy.RagnaDemoAuthentication][], which requires a matching username and
# password.

username = password = "Ragna"

response = client.post(
    "/token",
    data={"username": username, "password": password},
).raise_for_status()
token = response.json()

# %%
# We set the API token on our HTTP client so we don't have to manually supply it for
# each request below.

client.headers["Authorization"] = f"Bearer {token}"


# %%
# ## Step 3: Uploading documents
#
# Before we start with the upload process, let's first have a look what kind of
# documents are supported.

import json

response = client.get("/components").raise_for_status()
print(json.dumps(response.json(), indent=2))

# %%
# For simplicity, let's use a demo document with some information about Ragna

document_name = "ragna.txt"

with open(documentation_helpers.assets / document_name, "rb") as file:
    content = file.read()

print(content.decode())

# %%
# The upload process in Ragna consists of two parts:
#
# 1. Announce the file to be uploaded. Under the hood this pre-registers the document
#    in Ragnas database and returns information about how the upload is to be performed.
#    This is handled by the [ragna.core.Document][] class. By default,
#    [ragna.core.LocalDocument][] is used, which uploads the files to the local file
#    system.
# 2. Perform the actual upload with the information from step 1.

response = client.post("/document", json={"name": document_name}).raise_for_status()
document_upload = response.json()
print(json.dumps(response.json(), indent=2))

# %%
# The returned JSON contains two parts: the document object that we are later going to
# use to create a chat as well as the upload parameters.
# !!! note
#
#     The `"token"` in the response is *not* the Ragna REST API token, but rather a
#     separate one to perform the document upload.
#
# We perform the actual upload with the latter now.

document = document_upload["document"]

parameters = document_upload["parameters"]
client.request(
    parameters["method"],
    parameters["url"],
    data=parameters["data"],
    files={"file": content},
).raise_for_status()

# %%
# ## Step 4: Select a source storage and assistant
#
# The configuration we are using only supports demo components for the source storage
# and assistant and so we pick them here.

from ragna import source_storages, assistants

source_storage = source_storages.RagnaDemoSourceStorage.display_name()
assistant = assistants.RagnaDemoAssistant.display_name()

print(f"{source_storage=}, {assistant=}")

# %%
# ## Step 5: Start chatting
#
# Now that we have uploaded a document, and selected a source storage and assistant to
# be used, we can create a new chat.

response = client.post(
    "/chats",
    json={
        "name": "Tutorial REST API",
        "documents": [document],
        "source_storage": source_storage,
        "assistant": assistant,
        "params": {},
    },
).raise_for_status()
chat = response.json()
print(json.dumps(chat, indent=2))

# %%
# As can be seen by the `"prepared"` field in the `chat` JSON object we still need to
# prepare it.

client.post(f"/chats/{chat['id']}/prepare").raise_for_status()

# %%
# Finally, we can get answers to our questions.

response = client.post(
    f"/chats/{chat['id']}/answer",
    json={"prompt": "What is Ragna?"},
).raise_for_status()
answer = response.json()
print(json.dumps(answer, indent=2))

# %%

print(answer["content"])

# %%
# Before we close the tutorial, let's stop the REST API and have a look at what would
# have printed in the terminal if we had started it the regular way.

rest_api.stop()



import asyncio
import functools

import pytest

from ragna.core import Message


def sync(async_test_fn):
    @functools.wraps(async_test_fn)
    def wrapper(*args, **kwargs):
        return asyncio.run(async_test_fn(*args, **kwargs))

    return wrapper


class TestMessage:
    def test_fixed_content(self):
        content = "content"
        message = Message(content)

        assert message.content == content
        assert str(message) == content

    @sync
    async def test_fixed_content_read(self):
        content = "content"
        message = Message(content)

        assert (await message.read()) == content

    @sync
    async def test_fixed_content_iter(self):
        content = "content"
        message = Message(content)

        chunks = []
        async for chunk in message:
            chunks.append(chunk)
        assert chunks == [content]

    def make_content_stream(self, *chunks):
        async def content_stream():
            for chunk in chunks:
                yield chunk

        return content_stream()

    @pytest.mark.parametrize(
        "content_access",
        [
            pytest.param(lambda message: message.content, id="property"),
            str,
            repr,
        ],
    )
    def test_stream_content_access_error(self, content_access):
        content = "content"
        message = Message(self.make_content_stream(*content))

        with pytest.raises(RuntimeError):
            content_access(message)

    @sync
    async def test_stream_content_iter(self):
        content = "content"
        message = Message(self.make_content_stream(*content))

        chunks = []
        async for chunk in message:
            chunks.append(chunk)
        assert chunks == list(content)

        assert message.content == content

    @sync
    async def test_stream_content_read(self):
        content = "content"
        message = Message(self.make_content_stream(*content))

        assert (await message.read()) == content

        assert message.content == content



import json
from typing import AsyncIterator, cast

from ragna.core import RagnaException, Source

from ._api import ApiAssistant


class CohereApiAssistant(ApiAssistant):
    _API_KEY_ENV_VAR = "COHERE_API_KEY"
    _MODEL: str

    @classmethod
    def display_name(cls) -> str:
        return f"Cohere/{cls._MODEL}"

    def _make_preamble(self) -> str:
        return (
            "You are a helpful assistant that answers user questions given the included context. "
            "If you don't know the answer, just say so. Don't try to make up an answer. "
            "Only use the included documents below to generate the answer."
        )

    def _make_source_documents(self, sources: list[Source]) -> list[dict[str, str]]:
        return [{"title": source.id, "snippet": source.content} for source in sources]

    async def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> AsyncIterator[str]:
        # See https://docs.cohere.com/docs/cochat-beta
        # See https://docs.cohere.com/reference/chat
        # See https://docs.cohere.com/docs/retrieval-augmented-generation-rag
        async with self._client.stream(
            "POST",
            "https://api.cohere.ai/v1/chat",
            headers={
                "accept": "application/json",
                "content-type": "application/json",
                "authorization": f"Bearer {self._api_key}",
            },
            json={
                "preamble_override": self._make_preamble(),
                "message": prompt,
                "model": self._MODEL,
                "stream": True,
                "temperature": 0.0,
                "max_tokens": max_new_tokens,
                "documents": self._make_source_documents(sources),
            },
        ) as response:
            await self._assert_api_call_is_success(response)

            async for chunk in response.aiter_lines():
                event = json.loads(chunk)
                if event["event_type"] == "stream-end":
                    if event["event_type"] == "COMPLETE":
                        break

                    raise RagnaException(event["error_message"])
                if "text" in event:
                    yield cast(str, event["text"])


class Command(CohereApiAssistant):
    """
    [Cohere Command](https://docs.cohere.com/docs/models#command)

    !!! info "Required environment variables"

        - `COHERE_API_KEY`
    """

    _MODEL = "command"


class CommandLight(CohereApiAssistant):
    """
    [Cohere Command-Light](https://docs.cohere.com/docs/models#command)

    !!! info "Required environment variables"

        - `COHERE_API_KEY`
    """

    _MODEL = "command-light"



import asyncio

import panel as pn
import param

from . import js
from .central_view import CentralView
from .left_sidebar import LeftSidebar
from .modal_configuration import ModalConfiguration
from .modal_welcome import ModalWelcome
from .right_sidebar import RightSidebar


class MainPage(pn.viewable.Viewer, param.Parameterized):
    current_chat_id = param.String(default=None)
    chats = param.List(default=None)

    def __init__(self, api_wrapper, template):
        super().__init__()
        self.api_wrapper = api_wrapper
        self.template = template

        self.modal = None
        self.central_view = CentralView(api_wrapper=self.api_wrapper)
        self.central_view.on_click_chat_info = (
            lambda event, title, content: self.show_right_sidebar(title, content)
        )

        self.left_sidebar = LeftSidebar(api_wrapper=self.api_wrapper)
        self.left_sidebar.on_click_chat = self.on_click_chat
        self.left_sidebar.on_click_new_chat = lambda event: self.open_modal()

        self.right_sidebar = RightSidebar()

        pn.state.location.sync(
            self,
            {"current_chat_id": "current_chat_id"},
            on_error=lambda x: print(f"error sync on {x}"),
        )

    async def refresh_data(self):
        self.chats = await self.api_wrapper.get_chats()

    @param.depends("chats", watch=True)
    def after_update_chats(self):
        self.left_sidebar.chats = self.chats

        if len(self.chats) > 0:
            chat_id_exist = (
                len([c["id"] for c in self.chats if c["id"] == self.current_chat_id])
                > 0
            )

            if self.current_chat_id is None or not chat_id_exist:
                self.current_chat_id = self.chats[0]["id"]

            for c in self.chats:
                if c["id"] == self.current_chat_id:
                    self.central_view.set_current_chat(c)
                    break

    # Modal and callbacks
    def open_modal(self):
        self.modal = ModalConfiguration(
            api_wrapper=self.api_wrapper,
            new_chat_ready_callback=self.open_new_chat,
            cancel_button_callback=self.on_click_cancel_button,
        )

        self.template.modal.objects[0].objects = [self.modal]
        self.template.open_modal()

    def open_welcome_modal(self, event):
        self.modal = ModalWelcome(
            close_button_callback=lambda: self.template.close_modal(),
        )

        self.template.modal.objects[0].objects = [self.modal]
        self.template.open_modal()

    async def open_new_chat(self, new_chat_id):
        # called after creating a new chat.
        self.current_chat_id = new_chat_id
        await self.refresh_data()

        self.template.close_modal()

    def on_click_cancel_button(self, event):
        self.template.close_modal()

    # Left sidebar callbacks
    def on_click_chat(self, chat):
        self.central_view.set_loading(True)
        self.current_chat_id = chat["id"]
        self.central_view.set_current_chat(chat)
        self.central_view.set_loading(False)

    # Right sidebar callbacks
    def show_right_sidebar(self, title, content):
        self.right_sidebar.title = title
        self.right_sidebar.content = content
        self.right_sidebar.param.trigger("content")
        self.right_sidebar.show()

    @param.depends("current_chat_id", watch=True)
    def update_subviews_current_chat_id(self, avoid_senders=[]):
        if self.left_sidebar is not None and self.left_sidebar not in avoid_senders:
            self.left_sidebar.current_chat_id = self.current_chat_id

    def __panel__(self):
        asyncio.ensure_future(self.refresh_data())

        objects = [self.left_sidebar, self.central_view, self.right_sidebar]

        if self.chats is not None and len(self.chats) == 0:
            """I haven't found a better way to open the modal when the pages load,
            than simulating a click on the "New chat" button.
            - calling self.template.open_modal() doesn't work
            - calling self.on_click_new_chat doesn't work either
            - trying to schedule a call to on_click_new_chat with pn.state.schedule_task
                could have worked but my tests were yielding an unstable result.
            """

            new_chat_button_name = "open welcome modal"
            open_welcome_modal = pn.widgets.Button(
                name=new_chat_button_name,
                button_type="primary",
            )
            open_welcome_modal.on_click(self.open_welcome_modal)

            hack_open_modal = pn.pane.HTML(
                """
                            <script>   let buttons = $$$('button.bk-btn-primary');
                            console.log(buttons);
                                        buttons.forEach(function(btn){
                                            if ( btn.innerText.trim() == '{new_chat_btn_name}' ){
                                                btn.click();
                                            }
                                        });
                            </script>
                            """.replace(
                    "{new_chat_btn_name}", new_chat_button_name
                ).strip(),
                stylesheets=[":host { position:absolute; z-index:-999; }"],
            )

            objects.append(
                pn.Row(
                    open_welcome_modal,
                    pn.pane.HTML(js.SHADOWROOT_INDEXING),
                    hack_open_modal,
                    visible=False,
                )
            )

        main_page = pn.Row(
            *objects,
            stylesheets=[
                """   
                                :host { 
                                    background-color: rgb(248, 248, 248);
                                    height: 100%;
                                    width: 100%;
                                }

                                /* Enforces the width of the LeftSidebarn 
                                which is the "first of type" with this class 
                                (first object in the row) */
                                .bk-panel-models-layout-Column:first-of-type {
                                    min-width: 220px;
                                    width: 15%;     
                                }
                        """
            ],
        )

        return main_page



from typing import AsyncIterator

from ragna._compat import anext
from ragna.core import PackageRequirement, Requirement, Source

from ._api import ApiAssistant


# ijson does not support reading from an (async) iterator, but only from file-like
# objects, i.e. https://docs.python.org/3/tutorial/inputoutput.html#methods-of-file-objects.
# See https://github.com/ICRAR/ijson/issues/44 for details.
# ijson actually doesn't care about most of the file interface and only requires the
# read() method to be present.
class AsyncIteratorReader:
    def __init__(self, ait: AsyncIterator[bytes]) -> None:
        self._ait = ait

    async def read(self, n: int) -> bytes:
        # n is usually used to indicate how many bytes to read, but since we want to
        # return a chunk as soon as it is available, we ignore the value of n. The only
        # exception is n == 0, which is used by ijson to probe the return type and
        # set up decoding.
        if n == 0:
            return b""
        return await anext(self._ait, b"")  # type: ignore[call-arg]


class GoogleApiAssistant(ApiAssistant):
    _API_KEY_ENV_VAR = "GOOGLE_API_KEY"
    _MODEL: str

    @classmethod
    def _extra_requirements(cls) -> list[Requirement]:
        return [PackageRequirement("ijson")]

    @classmethod
    def display_name(cls) -> str:
        return f"Google/{cls._MODEL}"

    def _instructize_prompt(self, prompt: str, sources: list[Source]) -> str:
        # https://ai.google.dev/docs/prompt_best_practices#add-contextual-information
        return "\n".join(
            [
                "Answer the prompt using only the pieces of context below.",
                "If you don't know the answer, just say so. Don't try to make up additional context.",
                f"Prompt: {prompt}",
                *[f"\n{source.content}" for source in sources],
            ]
        )

    async def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> AsyncIterator[str]:
        import ijson

        async with self._client.stream(
            "POST",
            f"https://generativelanguage.googleapis.com/v1beta/models/{self._MODEL}:streamGenerateContent",
            params={"key": self._api_key},
            headers={"Content-Type": "application/json"},
            json={
                "contents": [
                    {"parts": [{"text": self._instructize_prompt(prompt, sources)}]}
                ],
                # https://ai.google.dev/docs/safety_setting_gemini
                "safetySettings": [
                    {"category": f"HARM_CATEGORY_{category}", "threshold": "BLOCK_NONE"}
                    for category in [
                        "HARASSMENT",
                        "HATE_SPEECH",
                        "SEXUALLY_EXPLICIT",
                        "DANGEROUS_CONTENT",
                    ]
                ],
                # https://ai.google.dev/tutorials/rest_quickstart#configuration
                "generationConfig": {
                    "temperature": 0.0,
                    "maxOutputTokens": max_new_tokens,
                },
            },
        ) as response:
            await self._assert_api_call_is_success(response)

            async for chunk in ijson.items(
                AsyncIteratorReader(response.aiter_bytes(1024)),
                "item.candidates.item.content.parts.item.text",
            ):
                yield chunk


class GeminiPro(GoogleApiAssistant):
    """[Google Gemini Pro](https://ai.google.dev/models/gemini)

    !!! info "Required environment variables"

        - `GOOGLE_API_KEY`

    !!! info "Required packages"

        - `ijson`
    """

    _MODEL = "gemini-pro"


class GeminiUltra(GoogleApiAssistant):
    """[Google Gemini Ultra](https://ai.google.dev/models/gemini)

    !!! info "Required environment variables"

        - `GOOGLE_API_KEY`

    !!! info "Required packages"

        - `ijson`
    """

    _MODEL = "gemini-ultra"



import abc
import contextlib
import json
import os
from typing import AsyncIterator

import httpx
from httpx import Response

import ragna
from ragna.core import Assistant, EnvVarRequirement, RagnaException, Requirement, Source


class ApiAssistant(Assistant):
    _API_KEY_ENV_VAR: str

    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [EnvVarRequirement(cls._API_KEY_ENV_VAR), *cls._extra_requirements()]

    @classmethod
    def _extra_requirements(cls) -> list[Requirement]:
        return []

    def __init__(self) -> None:
        self._client = httpx.AsyncClient(
            headers={"User-Agent": f"{ragna.__version__}/{self}"},
            timeout=60,
        )
        self._api_key = os.environ[self._API_KEY_ENV_VAR]

    async def answer(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int = 256
    ) -> AsyncIterator[str]:
        async for chunk in self._call_api(  # type: ignore[attr-defined, misc]
            prompt, sources, max_new_tokens=max_new_tokens
        ):
            yield chunk

    @abc.abstractmethod
    async def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> AsyncIterator[str]:
        ...

    async def _assert_api_call_is_success(self, response: Response) -> None:
        if response.is_success:
            return

        content = await response.aread()
        with contextlib.suppress(Exception):
            content = json.loads(content)

        raise RagnaException(
            "API call failed",
            request_method=response.request.method,
            request_url=str(response.request.url),
            response_status_code=response.status_code,
            response_content=content,
        )



from __future__ import annotations

import abc
import enum
import functools
import inspect
from typing import AsyncIterable, AsyncIterator, Iterator, Optional, Type, Union

import pydantic
import pydantic.utils

from ._document import Document
from ._utils import RequirementsMixin, merge_models


class Component(RequirementsMixin):
    """Base class for RAG components.

    !!! tip See also

        - [ragna.core.SourceStorage][]
        - [ragna.core.Assistant][]
    """

    @classmethod
    def display_name(cls) -> str:
        """
        Returns:
            Component name.
        """
        return cls.__name__

    def __repr__(self) -> str:
        return self.display_name()

    # FIXME: rename this to reflect that these methods can be parametrized from the chat
    #  level
    __ragna_protocol_methods__: list[str]

    @classmethod
    @functools.cache
    def _protocol_models(
        cls,
    ) -> dict[tuple[Type[Component], str], Type[pydantic.BaseModel]]:
        protocol_cls, protocol_methods = next(
            (cls_, cls_.__ragna_protocol_methods__)  # type: ignore[attr-defined]
            for cls_ in cls.__mro__
            if "__ragna_protocol_methods__" in cls_.__dict__
        )
        models = {}
        for method_name in protocol_methods:
            method = getattr(cls, method_name)
            concrete_params = inspect.signature(method).parameters
            protocol_params = inspect.signature(
                getattr(protocol_cls, method_name)
            ).parameters
            extra_param_names = concrete_params.keys() - protocol_params.keys()

            models[(cls, method_name)] = pydantic.create_model(  # type: ignore[call-overload]
                f"{cls.__name__}.{method_name}",
                **{
                    (param := concrete_params[param_name]).name: (
                        param.annotation,
                        param.default
                        if param.default is not inspect.Parameter.empty
                        else ...,
                    )
                    for param_name in extra_param_names
                },
            )
        return models

    @classmethod
    @functools.cache
    def _protocol_model(cls) -> Type[pydantic.BaseModel]:
        return merge_models(cls.display_name(), *cls._protocol_models().values())


class Source(pydantic.BaseModel):
    """Data class for sources stored inside a source storage.

    Attributes:
        id: Unique ID of the source.
        document: Document this source belongs to.
        location: Location of the source inside the document.
        content: Content of the source.
        num_tokens: Number of tokens of the content.
    """

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    id: str
    document: Document
    location: str
    content: str
    num_tokens: int


class SourceStorage(Component, abc.ABC):
    __ragna_protocol_methods__ = ["store", "retrieve"]

    @abc.abstractmethod
    def store(self, documents: list[Document]) -> None:
        """Store content of documents.

        Args:
            documents: Documents to store.
        """
        ...

    @abc.abstractmethod
    def retrieve(self, documents: list[Document], prompt: str) -> list[Source]:
        """Retrieve sources for a given prompt.

        Args:
            documents: Documents to retrieve sources from.
            prompt: Prompt to retrieve sources for.

        Returns:
            Matching sources for the given prompt ordered by relevance.
        """
        ...


class MessageRole(enum.Enum):
    """Message role

    Attributes:
        SYSTEM: The message was produced by the system. This includes the welcome
            message when [preparing a new chat][ragna.core.Chat.prepare] as well as
            error messages.
        USER: The message was produced by the user.
        ASSISTANT: The message was produced by an assistant.
    """

    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"


class Message:
    """Data class for messages.

    Attributes:
        role: The message producer.
        sources: The sources used to produce the message.

    !!! tip "See also"

        - [ragna.core.Chat.prepare][]
        - [ragna.core.Chat.answer][]
    """

    def __init__(
        self,
        content: Union[str, AsyncIterable[str]],
        *,
        role: MessageRole = MessageRole.SYSTEM,
        sources: Optional[list[Source]] = None,
    ) -> None:
        if isinstance(content, str):
            self._content: str = content
        else:
            self._content_stream: AsyncIterable[str] = content

        self.role = role
        self.sources = sources or []

    async def __aiter__(self) -> AsyncIterator[str]:
        if hasattr(self, "_content"):
            yield self._content
            return

        chunks = []
        async for chunk in self._content_stream:
            chunks.append(chunk)
            yield chunk

        self._content = "".join(chunks)

    async def read(self) -> str:
        if not hasattr(self, "_content"):
            # Since self.__aiter__ is already setting the self._content attribute, we
            # only need to exhaust the content stream here.
            async for _ in self:
                pass
        return self._content

    @property
    def content(self) -> str:
        if not hasattr(self, "_content"):
            raise RuntimeError(
                "Message content cannot be accessed without having iterated over it, "
                "e.g. `async for chunk in message`, or reading the content, e.g. "
                "`await message.read()`, first."
            )
        return self._content

    def __str__(self) -> str:
        return self.content

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"content={self.content}, role={self.role}, sources={self.sources}"
            f")"
        )


class Assistant(Component, abc.ABC):
    """Abstract base class for assistants used in [ragna.core.Chat][]"""

    __ragna_protocol_methods__ = ["answer"]

    @abc.abstractmethod
    def answer(self, prompt: str, sources: list[Source]) -> Iterator[str]:
        """Answer a prompt given some sources.

        Args:
            prompt: Prompt to be answered.
            sources: Sources to use when answering answer the prompt.

        Returns:
            Answer.
        """
        ...



import dataclasses
import itertools
from collections import deque
from typing import (
    Deque,
    Iterable,
    Iterator,
    Optional,
    TypeVar,
    cast,
)

from ragna._compat import itertools_pairwise
from ragna.core import (
    PackageRequirement,
    Page,
    Requirement,
    Source,
    SourceStorage,
)

T = TypeVar("T")


# The function is adapted from more_itertools.windowed to allow a ragged last window
# https://more-itertools.readthedocs.io/en/stable/api.html#more_itertools.windowed
def _windowed_ragged(
    iterable: Iterable[T], *, n: int, step: int
) -> Iterator[tuple[T, ...]]:
    window: Deque[T] = deque(maxlen=n)
    i = n
    for _ in map(window.append, iterable):
        i -= 1
        if not i:
            i = step
            yield tuple(window)

    if len(window) < n:
        yield tuple(window)
    elif 0 < i < min(step, n):
        yield tuple(window)[i:]


@dataclasses.dataclass
class Chunk:
    text: str
    page_numbers: Optional[list[int]]
    num_tokens: int


class VectorDatabaseSourceStorage(SourceStorage):
    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [
            # This looks like this should only be a requirement for Chroma, but it is
            # not. Chroma solved one major UX painpoint of vector DBs: the need for an
            # embedding function. Normally, one would pull in a massive amount
            # (in numbers as well as in size) of transitive dependencies that are hard
            # to manage and mostly not even used by the vector DB. Chroma provides a
            # wrapper around a compiled embedding function that has only minimal
            # requirements. We use this as base for all of our Vector DBs.
            PackageRequirement("chromadb>=0.4.13"),
            PackageRequirement("tiktoken"),
        ]

    def __init__(self) -> None:
        import chromadb.api
        import chromadb.utils.embedding_functions
        import tiktoken

        self._embedding_function = cast(
            chromadb.api.types.EmbeddingFunction,
            chromadb.utils.embedding_functions.DefaultEmbeddingFunction(),
        )
        # https://huggingface.co/sentence-transformers/all-MiniLM-L6-v2#all-minilm-l6-v2
        self._embedding_dimensions = 384
        self._tokenizer = tiktoken.get_encoding("cl100k_base")

    def _chunk_pages(
        self, pages: Iterable[Page], *, chunk_size: int, chunk_overlap: int
    ) -> Iterator[Chunk]:
        for window in _windowed_ragged(
            (
                (tokens, page.number)
                for page in pages
                for tokens in self._tokenizer.encode(page.text)
            ),
            n=chunk_size,
            step=chunk_size - chunk_overlap,
        ):
            tokens, page_numbers = zip(*window)
            yield Chunk(
                text=self._tokenizer.decode(tokens),  # type: ignore[arg-type]
                page_numbers=list(filter(lambda n: n is not None, page_numbers))
                or None,
                num_tokens=len(tokens),
            )

    def _page_numbers_to_str(self, page_numbers: Optional[Iterable[int]]) -> str:
        if not page_numbers:
            return ""

        page_numbers = sorted(set(page_numbers))
        if len(page_numbers) == 1:
            return str(page_numbers[0])

        ranges_str = []
        range_int = []
        for current_page_number, next_page_number in itertools_pairwise(
            itertools.chain(sorted(page_numbers), [None])
        ):
            current_page_number = cast(int, current_page_number)

            range_int.append(current_page_number)
            if next_page_number is None or next_page_number > current_page_number + 1:
                ranges_str.append(
                    ", ".join(map(str, range_int))
                    if len(range_int) < 3
                    else f"{range_int[0]}-{range_int[-1]}"
                )
                range_int = []

        return ", ".join(ranges_str)

    def _take_sources_up_to_max_tokens(
        self, sources: Iterable[Source], *, max_tokens: int
    ) -> list[Source]:
        taken_sources = []
        total = 0
        for source in sources:
            new_total = total + source.num_tokens
            if new_total > max_tokens:
                break

            taken_sources.append(source)
            total = new_total

        return taken_sources



import json
import os
import re
import subprocess

from mkdocs.plugins import get_plugin_logger

logger = get_plugin_logger(__file__)

# https://docs.readthedocs.io/en/stable/reference/environment-variables.html
RTD_ENV_VARS = {
    (name := f"READTHEDOCS{postfix}"): os.environ.get(name)
    for postfix in [
        "",
        "_PROJECT",
        "_LANGUAGE",
        "_VERSION",
        "_VERSION_NAME",
        "_VERSION_TYPE",
        "_VIRTUALENV_PATH",
        "_OUTPUT",
        "_CANONICAL_URL",
        "_GIT_CLONE_URL",
        "_GIT_IDENTIFIER",
        "_GIT_COMMIT_HASH",
    ]
}

RTD = RTD_ENV_VARS["READTHEDOCS"] == "True"


def git(*args):
    return subprocess.run(["git", *args], capture_output=True).stdout.decode()


def on_startup(command, dirty):
    if not RTD:
        return

    for name, value in RTD_ENV_VARS.items():
        logger.info(f"{name}={value}")

    # RTD makes some modifications to existing files and adds some new ones
    logger.info(
        json.dumps(
            {
                "files": re.findall(r"add '(.*?)'\n", git("add", "--dry-run", ".")),
                "diff": git("diff"),
            }
        )
    )



import docx
import pptx

from ragna.core import DocxDocumentHandler, LocalDocument, PptxDocumentHandler


def get_docx_document(tmp_path, docx_text):
    document = docx.Document()
    document.add_heading(docx_text)
    document.add_paragraph(docx_text)
    path = tmp_path / "test_document.docx"
    document.save(path)
    return LocalDocument.from_path(path)


def test_docx(tmp_path):
    docx_text = "ragna is neat!"
    tmp_docx_document = get_docx_document(tmp_path, docx_text)
    assert isinstance(tmp_docx_document.handler, DocxDocumentHandler)
    pages = list(tmp_docx_document.extract_pages())
    assert len(pages) == 2
    for page in pages:
        assert page.text == docx_text


def get_pptx_document(tmp_path, pptx_text):
    document = pptx.Presentation()
    document.slides.add_slide(document.slide_layouts[0])
    document.slides[0].shapes.title.text = pptx_text
    document.slides.add_slide(document.slide_layouts[0])
    document.slides[1].shapes.add_textbox(0, 0, 100, 100).text = pptx_text
    path = tmp_path / "test_document.pptx"
    document.save(path)
    return LocalDocument.from_path(path)


def test_pptx(tmp_path):
    pptx_text = "ragna is neat!"
    tmp_pptx_document = get_pptx_document(tmp_path, pptx_text)
    assert isinstance(tmp_pptx_document.handler, PptxDocumentHandler)
    pages = list(tmp_pptx_document.extract_pages())
    assert len(pages) == 2
    for page in pages:
        assert page.text == pptx_text



import os

from fastapi.testclient import TestClient

from ragna.core._utils import default_user


def authenticate(client: TestClient) -> None:
    username = default_user()
    token = (
        client.post(
            "/token",
            data={
                "username": username,
                "password": os.environ.get(
                    "RAGNA_DEMO_AUTHENTICATION_PASSWORD", username
                ),
            },
        )
        .raise_for_status()
        .json()
    )
    client.headers["Authorization"] = f"Bearer {token}"



from .styles import CONFIG_MODAL_MAX_HEIGHT, CONFIG_MODAL_MIN_HEIGHT


# Reset modal size
def reset_modal_size(width, height):
    return f"""
    document.getElementById("pn-Modal").style.setProperty("--dialog-height","{height}px", "important")
    document.getElementById("pn-Modal").style.setProperty("--dialog-width","{width}px", "important")
    """


JS_TOGGLE_CARD = f"""
let height = card.collapsed ? "{CONFIG_MODAL_MAX_HEIGHT}px" : "{CONFIG_MODAL_MIN_HEIGHT}px";
document.getElementById("pn-Modal").style.setProperty("--dialog-height",height, "important")
"""


# Allow selecting elements of the shadowroot with "$$$"
# https://discourse.holoviz.org/t/sync-scrollbars-for-wide-tabulators/5459
SHADOWROOT_INDEXING = """
<script type="text/javascript">
    function $$$(selector, rootNode=document.body) {
        const arr = []

        const traverser = node => {
            // 1. decline all nodes that are not elements
            if(node.nodeType !== Node.ELEMENT_NODE) {
                return
            }

            // 2. add the node to the array, if it matches the selector
            if(node.matches(selector)) {
                arr.push(node)
            }

            // 3. loop through the children
            const children = node.children
            if (children.length) {
                for(const child of children) {
                    traverser(child)
                }
            }

            // 4. check for shadow DOM, and loop through it's children
            const shadowRoot = node.shadowRoot
            if (shadowRoot) {
                const shadowChildren = shadowRoot.children
                for(const shadowChild of shadowChildren) {
                    traverser(shadowChild)
                }
            }
        }

        traverser(rootNode)
        return arr
    }
</script>
""".strip()


""" This is a fix to close the modal only when click *down* on it's outside area.
We have seen a bug where, if you start by clicking down somewhere inside and end click up
outside of it (like selectiong the ChatName text input with the mouse ),
it's considered as a click outside the modal so it needs to be closed.
"""
MODAL_MOUSE_UP_FIX = """
<script>
    window.onclick = null

    window.onmousedown = (event) => {
        if ( event.target == modal){
            modal.style.display = "none";
        }
    };
</script>
""".strip()

CONNECTION_MONITOR = """
<script>
    const originalSend = WebSocket.prototype.send;
    window.sockets = [];
    WebSocket.prototype.send = function(...args) {
        if (window.sockets.indexOf(this) === -1)
            window.sockets.push(this);
        return originalSend.call(this, ...args);
    };

    const polling = setInterval(function() {

        if ( window.sockets.length > 0 ){

            if ( window.sockets[0].readyState >= 2 ){

                let div = document.createElement('div');
                div.style.color = 'white';
                div.style.backgroundColor= 'crimson';
                div.style.padding = '10px 10px 10px 10px';
                div.style.textAlign= 'center';

                let text = document.createTextNode('Session has expired. Please reload.');
                div.appendChild(text);


                window.document.body.insertBefore(
                    div,
                    window.document.body.firstChild
                );

                clearInterval(polling);
            }
        }

    }, 5000);

</script>
"""



import inspect
import os
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Optional

import httpx

from ragna._utils import timeout_after
from ragna.deploy import Config

__all__ = ["assets", "RestApi"]

assets = Path(__file__).parent / "assets"


class RestApi:
    def __init__(self):
        self._process: Optional[subprocess.Popen] = None

    def start(self, config: Config, *, authenticate: bool = False) -> httpx.Client:
        python_path, config_path = self._prepare_config(config)

        client = httpx.Client(base_url=config.api.url)

        self._process = self._start_api(config_path, python_path, client)

        if authenticate:
            self._authenticate(client)

        return client

    def _prepare_config(self, config: Config) -> tuple[str, str]:
        deploy_directory = Path(tempfile.mkdtemp())

        python_path = (
            f"{deploy_directory}{os.pathsep}{os.environ.get('PYTHONPATH', '')}"
        )
        config_path = str(deploy_directory / "ragna.toml")

        config.local_root = deploy_directory

        sys.modules["__main__"].__file__ = inspect.getouterframes(
            inspect.currentframe()
        )[2].filename
        custom_module = deploy_directory.name
        with open(deploy_directory / f"{custom_module}.py", "w") as file:
            # TODO: this currently only handles assistants. When needed, we can extend
            #  to source storages.
            file.write("from ragna import assistants\n\n")

            for assistant in config.assistants:
                if assistant.__module__ == "__main__":
                    file.write(f"{inspect.getsource(assistant)}\n\n")
                    assistant.__module__ = custom_module

        config.to_file(config_path)
        return python_path, config_path

    def _start_api(
        self, config_path: str, python_path: str, client: httpx.Client
    ) -> subprocess.Popen:
        env = os.environ.copy()
        env["PYTHONPATH"] = python_path

        process = subprocess.Popen(
            [sys.executable, "-m", "ragna", "api", "--config", config_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )

        def check_api_available() -> bool:
            try:
                return client.get("/").is_success
            except httpx.ConnectError:
                return False

        failure_message = "Failed to the start the Ragna REST API."

        @timeout_after(60, message=failure_message)
        def wait_for_api() -> None:
            print("Starting Ragna REST API")
            while not check_api_available():
                try:
                    stdout, stderr = process.communicate(timeout=1)
                except subprocess.TimeoutExpired:
                    print(".", end="")
                    continue
                else:
                    parts = [failure_message]
                    if stdout:
                        parts.append(f"\n\nSTDOUT:\n\n{stdout.decode()}")
                    if stderr:
                        parts.append(f"\n\nSTDERR:\n\n{stderr.decode()}")

                    raise RuntimeError("".join(parts))

            print()

        wait_for_api()
        return process

    def _authenticate(self, client: httpx.Client) -> None:
        username = password = "Ragna"

        response = client.post(
            "/token",
            data={"username": username, "password": password},
        ).raise_for_status()
        token = response.json()

        client.headers["Authorization"] = f"Bearer {token}"

    def stop(self, *, quiet: bool = False) -> None:
        self._process.kill()
        stdout, _ = self._process.communicate()

        if not quiet:
            print(stdout.decode())

    def __del__(self):
        if self._process is not None:
            self.stop(quiet=True)



import os
from pathlib import Path
from urllib.parse import urlsplit

import pytest

from ragna.core import RagnaException
from ragna.deploy import Config


def test_env_var_prefix(mocker, tmp_path):
    env_var = tmp_path / "env_var"
    mocker.patch.dict(os.environ, values={"RAGNA_LOCAL_ROOT": str(env_var)})

    config = Config()

    assert config.local_root == env_var


def test_env_var_api_prefix(mocker):
    env_var = "hostname"
    mocker.patch.dict(os.environ, values={"RAGNA_API_HOSTNAME": env_var})

    config = Config()

    assert config.api.hostname == env_var


def test_env_var_ui_prefix(mocker):
    env_var = "hostname"
    mocker.patch.dict(os.environ, values={"RAGNA_UI_HOSTNAME": env_var})

    config = Config()

    assert config.ui.hostname == env_var


@pytest.mark.xfail()
def test_explicit_gt_env_var(mocker, tmp_path):
    explicit = tmp_path / "explicit"

    env_var = tmp_path / "env_var"
    mocker.patch.dict(os.environ, values={"RAGNA_LOCAL_ROOT": str(env_var)})

    config = Config(local_root=explicit)

    assert config.local_root == explicit


def test_env_var_gt_config_file(mocker, tmp_path):
    config_file = tmp_path / "config_file"
    config = Config(local_root=config_file)
    assert config.local_root == config_file

    config_path = tmp_path / "ragna.toml"
    config.to_file(config_path)

    env_var = tmp_path / "env_var"
    mocker.patch.dict(os.environ, values={"RAGNA_LOCAL_ROOT": str(env_var)})

    config = Config.from_file(config_path)

    assert config.local_root == env_var


def test_api_database_url_default_path(tmp_path):
    config = Config(local_root=tmp_path)
    assert Path(urlsplit(config.api.database_url).path[1:]).parent == tmp_path


@pytest.mark.parametrize("config_subsection", ["api", "ui"])
def test_origins_default(config_subsection):
    hostname, port = "0.0.0.0", "80"
    config = Config(ui=dict(hostname=hostname, port=port))

    assert getattr(config, config_subsection).origins == [f"http://{hostname}:{port}"]


def test_from_file_path_not_exists(tmp_path):
    with pytest.raises(RagnaException, match="does not exist"):
        Config.from_file(tmp_path / "ragna.toml")


def test_from_file_path_exists(tmp_path):
    config_path = tmp_path / "ragna.toml"
    open(config_path, "w").close()

    with pytest.raises(RagnaException, match="already exists"):
        Config().to_file(config_path)


def test_from_file_path_exists_force(tmp_path):
    config_path = tmp_path / "ragna.toml"
    open(config_path, "w").close()

    config = Config(local_root=tmp_path)
    config.to_file(config_path, force=True)

    assert Config.from_file(config_path).local_root == tmp_path



"""
# Streaming messages

Ragna supports streaming responses from the assistant. This example showcases how this
is performed using the Python and REST API.
"""

# %%
# Before we start this example, we import some helpers.

import sys
from pathlib import Path

sys.path.insert(0, str(Path.cwd().parent))

import documentation_helpers

# %%
# ## Setup streaming assistant
#
# To be able to stream a message from an assistant, it needs to support streaming. For
# this example, we subclass the [ragna.assistants.RagnaDemoAssistant][], split its
# message on whitespace, and return the individual chunks.
#
# !!! tip
#
#     Of the assistants that Ragna has built in, the following ones support streaming:
#
#     - [Anthropic](https://www.anthropic.com/)
#       - [ragna.assistants.Claude][]
#       - [ragna.assistants.ClaudeInstant][]
#     - [Cohere](https://cohere.com/)
#       - [ragna.assistants.Command][]
#       - [ragna.assistants.CommandLight][]
#     - [Google](https://ai.google.dev/)
#       - [ragna.assistants.GeminiPro][]
#       - [ragna.assistants.GeminiUltra][]
#     - [OpenAI](https://openai.com/)
#       - [ragna.assistants.Gpt35Turbo16k][]
#       - [ragna.assistants.Gpt4][]

from ragna import assistants


class DemoStreamingAssistant(assistants.RagnaDemoAssistant):
    def answer(self, prompt, sources):
        content = next(super().answer(prompt, sources))
        for chunk in content.split(" "):
            yield f"{chunk} "


# %%
# ## Python API
#
# Let's create and prepare a chat using the assistant we have defined above.

from ragna import Rag, source_storages

document_path = documentation_helpers.assets / "ragna.txt"

chat = Rag().chat(
    documents=[document_path],
    source_storage=source_storages.RagnaDemoSourceStorage,
    assistant=DemoStreamingAssistant,
)
_ = await chat.prepare()

# %%

message = await chat.answer("What is Ragna?", stream=True)

# %%
# At this stage, we cannot access the content of the message, e.g. by printing it.

try:
    print(message)
except Exception as error:
    print(f"{type(error).__name__}: {error}")

# %%
# To get the individual chunks, we asynchronously iterate over the `message`.

chunks = [chunk async for chunk in message]

print(len(chunks))
print(chunks)

# %%
# Joining the chunks together results in the full message.

print("".join(chunks))

# %%
# ## REST API

from ragna.deploy import Config

config = Config(assistants=[DemoStreamingAssistant])

rest_api = documentation_helpers.RestApi()

client = rest_api.start(config, authenticate=True)

# %%
# Upload the document.

document_upload = (
    client.post("/document", json={"name": document_path.name})
    .raise_for_status()
    .json()
)

document = document_upload["document"]

parameters = document_upload["parameters"]
client.request(
    parameters["method"],
    parameters["url"],
    data=parameters["data"],
    files={"file": open(document_path, "rb")},
).raise_for_status()

# %%
# Start and prepare the chat

chat = (
    client.post(
        "/chats",
        json={
            "name": "Tutorial REST API",
            "documents": [document],
            "source_storage": source_storages.RagnaDemoSourceStorage.display_name(),
            "assistant": DemoStreamingAssistant.display_name(),
            "params": {},
        },
    )
    .raise_for_status()
    .json()
)

client.post(f"/chats/{chat['id']}/prepare").raise_for_status()

# %%
# Streaming the response is performed with [JSONL](https://jsonlines.org/). Each line
# in the response is valid JSON and corresponds to one chunk.

import json


with client.stream(
    "POST",
    f"/chats/{chat['id']}/answer",
    json={"prompt": "What is Ragna?", "stream": True},
) as response:
    chunks = [json.loads(data) for data in response.iter_lines()]

# %%
# The first chunk contains the full message object including the sources along the first
# chunk of the content.

print(len(chunks))
print(json.dumps(chunks[0], indent=2))

# %%
# Subsequent chunks no longer contain the sources.

print(json.dumps(chunks[1], indent=2))

# %%
# Joining the content of the chunks together results in the full message.

print("".join(chunk["content"] for chunk in chunks))

# %%
# Before we close the example, let's stop the REST API and have a look at what would
# have printed in the terminal if we had started it the regular way.

rest_api.stop()



import contextlib
import uuid
from typing import Annotated, Any, AsyncIterator, Iterator, Type, cast

import aiofiles
from fastapi import (
    Body,
    Depends,
    FastAPI,
    Form,
    HTTPException,
    Request,
    UploadFile,
    status,
)
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, StreamingResponse
from pydantic import BaseModel

import ragna
import ragna.core
from ragna._compat import aiter, anext
from ragna._utils import handle_localhost_origins
from ragna.core import Assistant, Component, Rag, RagnaException, SourceStorage
from ragna.core._rag import SpecialChatParams
from ragna.deploy import Config

from . import database, schemas


def app(*, config: Config, ignore_unavailable_components: bool) -> FastAPI:
    ragna.local_root(config.local_root)

    rag = Rag()  # type: ignore[var-annotated]
    components_map: dict[str, Component] = {}
    for components in [config.source_storages, config.assistants]:
        components = cast(list[Type[Component]], components)
        at_least_one = False
        for component in components:
            loaded_component = rag._load_component(
                component, ignore_unavailable=ignore_unavailable_components
            )
            if loaded_component is None:
                print(
                    f"Ignoring {component.display_name()}, because it is not available."
                )
            else:
                at_least_one = True
                components_map[component.display_name()] = loaded_component

        if not at_least_one:
            raise RagnaException(
                "No component available",
                components=[component.display_name() for component in components],
            )

    def get_component(display_name: str) -> Component:
        component = components_map.get(display_name)
        if component is None:
            raise RagnaException(
                "Unknown component",
                display_name=display_name,
                http_status_code=status.HTTP_404_NOT_FOUND,
                http_detail=RagnaException.MESSAGE,
            )

        return component

    app = FastAPI(
        title="ragna",
        version=ragna.__version__,
        root_path=config.api.root_path,
    )
    app.add_middleware(
        CORSMiddleware,
        allow_origins=handle_localhost_origins(config.api.origins),
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    @app.exception_handler(RagnaException)
    async def ragna_exception_handler(
        request: Request, exc: RagnaException
    ) -> JSONResponse:
        if exc.http_detail is RagnaException.EVENT:
            detail = exc.event
        elif exc.http_detail is RagnaException.MESSAGE:
            detail = str(exc)
        else:
            detail = cast(str, exc.http_detail)
        return JSONResponse(
            status_code=exc.http_status_code,
            content={"error": {"message": detail}},
        )

    @app.get("/")
    async def version() -> str:
        return ragna.__version__

    authentication = config.authentication()

    @app.post("/token")
    async def create_token(request: Request) -> str:
        return await authentication.create_token(request)

    UserDependency = Annotated[str, Depends(authentication.get_user)]

    def _get_component_json_schema(
        component: Type[Component],
    ) -> dict[str, dict[str, Any]]:
        json_schema = component._protocol_model().model_json_schema()
        # FIXME: there is likely a better way to exclude certain fields builtin in
        #  pydantic
        for special_param in SpecialChatParams.model_fields:
            if (
                "properties" in json_schema
                and special_param in json_schema["properties"]
            ):
                del json_schema["properties"][special_param]
            if "required" in json_schema and special_param in json_schema["required"]:
                json_schema["required"].remove(special_param)
        return json_schema

    @app.get("/components")
    async def get_components(_: UserDependency) -> schemas.Components:
        return schemas.Components(
            documents=sorted(config.document.supported_suffixes()),
            source_storages=[
                _get_component_json_schema(type(source_storage))
                for source_storage in components_map.values()
                if isinstance(source_storage, SourceStorage)
            ],
            assistants=[
                _get_component_json_schema(type(assistant))
                for assistant in components_map.values()
                if isinstance(assistant, Assistant)
            ],
        )

    make_session = database.get_sessionmaker(config.api.database_url)

    @contextlib.contextmanager
    def get_session() -> Iterator[database.Session]:
        with make_session() as session:  # type: ignore[attr-defined]
            yield session

    @app.post("/document")
    async def create_document_upload_info(
        user: UserDependency,
        name: Annotated[str, Body(..., embed=True)],
    ) -> schemas.DocumentUpload:
        with get_session() as session:
            document = schemas.Document(name=name)
            metadata, parameters = await config.document.get_upload_info(
                config=config, user=user, id=document.id, name=document.name
            )
            database.add_document(
                session, user=user, document=document, metadata=metadata
            )
            return schemas.DocumentUpload(parameters=parameters, document=document)

    @app.put("/document")
    async def upload_document(
        token: Annotated[str, Form()], file: UploadFile
    ) -> schemas.Document:
        if not issubclass(config.document, ragna.core.LocalDocument):
            raise HTTPException(
                status_code=400,
                detail="Ragna configuration does not support local upload",
            )
        with get_session() as session:
            user, id = ragna.core.LocalDocument.decode_upload_token(token)
            document, metadata = database.get_document(session, user=user, id=id)

            core_document = ragna.core.LocalDocument(
                id=document.id, name=document.name, metadata=metadata
            )
            core_document.path.parent.mkdir(parents=True, exist_ok=True)
            async with aiofiles.open(core_document.path, "wb") as document_file:
                while content := await file.read(1024):
                    await document_file.write(content)

            return document

    def schema_to_core_chat(
        session: database.Session, *, user: str, chat: schemas.Chat
    ) -> ragna.core.Chat:
        core_chat = rag.chat(
            documents=[
                config.document(
                    id=document.id,
                    name=document.name,
                    metadata=database.get_document(
                        session,
                        user=user,
                        id=document.id,
                    )[1],
                )
                for document in chat.metadata.documents
            ],
            source_storage=get_component(chat.metadata.source_storage),  # type: ignore[arg-type]
            assistant=get_component(chat.metadata.assistant),  # type: ignore[arg-type]
            user=user,
            chat_id=chat.id,
            chat_name=chat.metadata.name,
            **chat.metadata.params,
        )
        # FIXME: We need to reconstruct the previous messages here. Right now this is
        #  not needed, because the chat itself never accesses past messages. However,
        #  if we implement a chat history feature, i.e. passing past messages to
        #  the assistant, this becomes crucial.
        core_chat._messages = []
        core_chat._prepared = chat.prepared

        return core_chat

    @app.post("/chats")
    async def create_chat(
        user: UserDependency,
        chat_metadata: schemas.ChatMetadata,
    ) -> schemas.Chat:
        with get_session() as session:
            chat = schemas.Chat(metadata=chat_metadata)

            # Although we don't need the actual ragna.core.Chat object here,
            # we use it to validate the documents and metadata.
            schema_to_core_chat(session, user=user, chat=chat)

            database.add_chat(session, user=user, chat=chat)
            return chat

    @app.get("/chats")
    async def get_chats(user: UserDependency) -> list[schemas.Chat]:
        with get_session() as session:
            return database.get_chats(session, user=user)

    @app.get("/chats/{id}")
    async def get_chat(user: UserDependency, id: uuid.UUID) -> schemas.Chat:
        with get_session() as session:
            return database.get_chat(session, user=user, id=id)

    @app.post("/chats/{id}/prepare")
    async def prepare_chat(user: UserDependency, id: uuid.UUID) -> schemas.Message:
        with get_session() as session:
            chat = database.get_chat(session, user=user, id=id)

            core_chat = schema_to_core_chat(session, user=user, chat=chat)

            welcome = schemas.Message.from_core(await core_chat.prepare())

            chat.prepared = True
            chat.messages.append(welcome)
            database.update_chat(session, user=user, chat=chat)

            return welcome

    @app.post("/chats/{id}/answer")
    async def answer(
        user: UserDependency,
        id: uuid.UUID,
        prompt: Annotated[str, Body(..., embed=True)],
        stream: Annotated[bool, Body(..., embed=True)] = False,
    ) -> schemas.Message:
        with get_session() as session:
            chat = database.get_chat(session, user=user, id=id)
            chat.messages.append(
                schemas.Message(content=prompt, role=ragna.core.MessageRole.USER)
            )
            core_chat = schema_to_core_chat(session, user=user, chat=chat)

        core_answer = await core_chat.answer(prompt, stream=stream)

        if stream:

            async def message_chunks() -> AsyncIterator[BaseModel]:
                core_answer_stream = aiter(core_answer)
                content_chunk = await anext(core_answer_stream)

                answer = schemas.Message(
                    content=content_chunk,
                    role=core_answer.role,
                    sources=[
                        schemas.Source.from_core(source)
                        for source in core_answer.sources
                    ],
                )
                yield answer

                # Avoid sending the sources multiple times
                answer_chunk = answer.model_copy(update=dict(sources=None))
                content_chunks = [answer_chunk.content]
                async for content_chunk in core_answer_stream:
                    content_chunks.append(content_chunk)
                    answer_chunk.content = content_chunk
                    yield answer_chunk

                with get_session() as session:
                    answer.content = "".join(content_chunks)
                    chat.messages.append(answer)
                    database.update_chat(session, user=user, chat=chat)

            async def to_jsonl(models: AsyncIterator[Any]) -> AsyncIterator[str]:
                async for model in models:
                    yield f"{model.model_dump_json()}\n"

            return StreamingResponse(  # type: ignore[return-value]
                to_jsonl(message_chunks())
            )
        else:
            answer = schemas.Message.from_core(core_answer)

            with get_session() as session:
                chat.messages.append(answer)
                database.update_chat(session, user=user, chat=chat)

            return answer

    @app.delete("/chats/{id}")
    async def delete_chat(user: UserDependency, id: uuid.UUID) -> None:
        with get_session() as session:
            database.delete_chat(session, user=user, id=id)

    return app
